{: XMLMap
	DataSet to XML mapper that can read/write datasets (master/details also) from/to XML

	Author: Tomislav Karda	
}

unit XMLMap;

interface

uses
	Windows, Classes, SysUtils, DB, XMLNode;

type
	TXMLMapFieldDirection = (xdReadOnly, xdWriteOnly, xdReadWrite);

type
	TXMLMap = class;

	TXMLMapField = class(TCollectionItem)
	private
		FFieldName: string;
		FNodeName: string;
		FDirection: TXMLMapFieldDirection;
		FWriteCDATA: boolean;
		FField: TField;

		procedure SetFieldName(Value: string);
		function IsNodeNameStored: boolean;
	protected
		function GetDisplayName: string; override;
		function IsFieldObjectStored: boolean;
	public
		constructor Create(Collection: TCollection); override;
		procedure Assign(Source: TPersistent); override;
	published
		(*
			FieldName:
							Name of dataset field that will be mapped to XML element node
			NodeName:
							Element node name that is mapped to dataset field
			Direction:
							Mapping direction. Field can be mapped only for reading, writing or both
			WriteCDATA:
							If set then field content is written as CDATA section in XML while write.
			FieldObject:
							Dataset field object that is mapped. This property is automaticaly set when
							FieldName is changed, but can be altered manualy.
		*)
		property FieldName: string read FFieldName write SetFieldName;
		property NodeName: string read FNodeName write FNodeName stored IsNodeNameStored;
		property Direction: TXMLMapFieldDirection read FDirection write FDirection default xdReadWrite;
		property WriteCDATA: boolean read FWriteCDATA write FWriteCDATA default false;
		property FieldObject: TField read FField write FField stored IsFieldObjectStored;
	end;

	TXMLMapFields = class(TOwnedCollection)
	private
		function GetItem(Index: Integer): TXMLMapField;
		procedure SetItem(Index: Integer; Value: TXMLMapField);
	protected
	public
		constructor Create(AOwner: TXMLMap);
		function Add: TXMLMapField;
		property Items[Index: Integer]: TXMLMapField read GetItem write SetItem; default;
		function Owner: TXMLMap;
	end;

	(*
	TXMLMapFieldRead = procedure (Sender: TObject; Field: TField; XMLData: string) of object;
	TXMLMapFieldWrite = procedure (Sender: TObject; Field: TField; var XMLData: string) of object;
	*)

  TXMLMapRecordReadAction = (xnaAppend, xnaEdit, xnaSkip);

	TXMLMapNotifyEvent = procedure (Sender: TObject; Node: TXMLNode) of object;
	TXMLMapOnRecordWriteEvent = procedure (Sender: TObject; var Write: boolean) of object;
	TXMLMapOnFieldWriteEvent = procedure (Sender: TObject; MapField: TXMLMapField; var Write: boolean) of object;
	TXMLMapOnRecordReadEvent = procedure (Sender: TObject; Node: TXMLNode; var Action: TXMLMapRecordReadAction) of object;
	TXMLMapOnFieldReadEvent = procedure (Sender: TObject; MapField: TXMLMapField; Node: TXMLNode; var Read: boolean) of object;

	TXMLMap = class(TComponent)
	private
		FMapFields: TXMLMapFields;
		FDataSet: TDataSet;
		FRecordNodeName: string;
		FRecordsNodeName: string;
		FDateFormat: string;
		FTimeFormat: string;
		FDecimalSeparator: char;
		FVersion: string;
		FMaster: TXMLMap;
		FDetails: TList;

		FBeforeDataSetWrite: TXMLMapNotifyEvent;
		FBeforeRecordWrite: TXMLMapNotifyEvent;
		FAfterRecordWrite: TXMLMapNotifyEvent;
		FAfterDataSetWrite: TXMLMapNotifyEvent;
		FOnRecordWrite: TXMLMapOnRecordWriteEvent;
		FOnFieldWrite: TXMLMapOnFieldWriteEvent;

		FBeforeDataSetRead: TXMLMapNotifyEvent;
		FBeforeRecordRead: TXMLMapNotifyEvent;
		FAfterRecordRead: TXMLMapNotifyEvent;
		FAfterDataSetRead: TXMLMapNotifyEvent;
		FOnRecordRead: TXMLMapOnRecordReadEvent;
		FOnFieldRead: TXMLMapOnFieldReadEvent;

		procedure SetMapFields(Value: TXMLMapFields);
		procedure SetDataSet(Value: TDataSet);
		procedure SetMaster(Value: TXMLMap);
	protected
		procedure DoBeforeDataSetWrite(Node: TXMLNode); virtual;
		procedure DoBeforeRecordWrite(Node: TXMLNode); virtual;
		procedure DoAfterRecordWrite(Node: TXMLNode); virtual;
		procedure DoAfterDataSetWrite(Node: TXMLNode); virtual;
		procedure DoOnRecordWrite(var Write: boolean); virtual;
		procedure DoOnFieldWrite(MapField: TXMLMapField; var Write: boolean); virtual;

		procedure DoBeforeDataSetRead(Node: TXMLNode); virtual;
		procedure DoBeforeRecordRead(Node: TXMLNode); virtual;
		procedure DoAfterRecordRead(Node: TXMLNode); virtual;
		procedure DoAfterDataSetRead(Node: TXMLNode); virtual;
		procedure DoOnRecordRead(Node: TXMLNode; var Action: TXMLMapRecordReadAction); virtual;
		procedure DoOnFieldRead(MapField: TXMLMapField; Node: TXMLNode; var Read: boolean); virtual;
	public
		constructor Create(AOwner: TComponent); override;
		destructor Destroy; override;

		procedure RegisterDetail(d: TXMLMap);
		procedure UnRegisterDetail(d: TXMLMap);

		(*
			WriteRecord:
							Creates new TXMLNode object and fills it with the current record content of
							dataset. If detail XMLMap components are linked to this XMLMap then also
							detail records are filled in. For every record WriteRecord is called to write
              a record. Default action is to write all records but this can be changed in
              OnRecordWrite event where application	can skip unwanted records.
			WriteRecords:
							Creates new TXMLNode object and fills it with the current records of
							dataset. For every record of dataset WriteRecord is called to produce
							node for this record. All records are gathered into one final xml node.
			ReadRecord:
							Reads current dataset record from given TXMLNode object. If detail XMLMap
							components are linked to this XMLMap then also detail records are read.
							Dataset must be in Edit or Insert state, otherwise exception will be raised.
							After the read dataset will stay in Edit or Insert mode.
			ReadRecords:
							Loads dataset records from given TXMLNode object. For every record
							ReadRecord is called to read a record. Default action is to append records
							while reading but this can be changed in OnRecordRead event where application
							can seek the appropriate record for current record node and tell the reader to
							edit and not append the record.
		*)
		function WriteRecord: TXMLNode;
		function WriteRecords: TXMLNode;
		procedure ReadRecord (Node: TXMLNode);
		procedure ReadRecords (Node: TXMLNode);
	published
		(*
			DataSet:
							DataSet which data records are to be mapped to XML document.
			MapFields:
							Individual dataset fields mapping to element nodes from XML document.
			RecordNodeName:
							Element node name that contains one record from/for DataSet data.
			RecordsNodeName:
							Element node name that contains set of records from/for DataSet data.
			DateFormat:
							Format that is used for storing Date part of TDateTime and TDate field values in
							xml element node data values.
			TimeFormat:
							Format that is used for storing Time part of TDateTime and TTime field values in
							xml element node data values.
			DecimalSeparator:
							Decimal separator that is used for representing numbers with point.
			Version:
							Version mark that is automaticaly writen to the records element node as an attribute.
							This is introduced so that different versions of the same dataset of data can be traced
							and converted by the application. So, version is automaticaly appended and if there may exist
							few versions, application can perform check and/or conversion in BeforeDataSetRead. There is no
							built in operation for version other than just writing the version attribute value.
			Master:
							Master property represents master TXMLMap if this TXMLMap is a mapping of detail dataset
							for a master dataset that is also mapped. In this way, in one processing of read or write
							on the level of master TXMLMap, automaticaly for every master record all detail records are
							also processed.
		*)
		property DataSet: TDataSet read FDataSet write SetDataSet;
		property MapFields: TXMLMapFields read FMapFields write SetMapFields;
		property RecordNodeName: string read FRecordNodeName write FRecordNodeName;
		property RecordsNodeName: string read FRecordsNodeName write FRecordsNodeName;
		property DateFormat: string read FDateFormat write FDateFormat;
		property TimeFormat: string read FTimeFormat write FTimeFormat;
		property DecimalSeparator: char read FDecimalSeparator write FDecimalSeparator;
		property Version: string read FVersion write FVersion;
		property Master: TXMLMap read FMaster write SetMaster;

		(*
			BeforeDataSetWrite:
							Occurs before the write of the dataset records to the XML element node. If this TXMLMap has a master
							TXMLMap property set then this event will be fired for every record in master dataset, and detail records
							will be written as a child node in a master record node.
			OnRecordWrite:
							Occurs before BeforeRecordWrite event, and is used to override default action when writing record into
							the xml node. Default action is to append the current record, but record can be skipped in writing
              if needed by seting Write to false.
			BeforeRecordWrite:
							Occurs before the write of the dataset record to the XML element node.
			OnFieldWrite:
							Occurs once for every field in the collection. Fields can be skipped in writing
							if needed by seting Write to false.
			AfterRecordWrite:
							Occurs after the write of the dataset record to the XML element node.
			AfterDataSetWrite:
							Occurs after the write of the dataset records to the XML element node. All records are written into
							the node as child node elements for every record.

			BeforeDataSetRead:
							Occurs before the read of the dataset records from the XML element node. If this TXMLMap has a master
							TXMLMap property set then this event will be fired for every record in master dataset that is found in a
							XML document, and detail records will be taken from a child node in a master record node.
			OnRecordRead:
							Occurs before BeforeRecordRead event, and is used to override default action when reading record into
							the dataset. Default action is to append the record that is read, but if there is a case of update
							of the existing record from the dataset then application can use this event to seek the appropriate
							record in the dataset and set Action flag to xnaEdit indicating that reading will be by Dataset.Edit method
							and not the Dataset.Append. By setting Action flag to xnaSkip this record will be skipped and not processed.
			BeforeRecordRead:
							Occurs before the read of the dataset record from the XML element node. Dataset is already put into the
							edit or insert mode according to OnRecordRead event.
			OnFieldRead:
							Occurs once for every field in the collection. Fields can be skipped in reading
							if needed by seting Read to false. Node can be nil, if this field is not in the xml.
			AfterRecordRead:
							Occurs after the read of the dataset record from the XML element node, and also after all the detail
							datasets have been read. Dataset is still in the insert or edit mode so the application can make
							additional changes to the dataset.
			AfterDataSetRead:
							Occurs after the read of the dataset records from the XML element node. All records are read into
							the dataset from the child element nodes of the node argument.
		*)
		property BeforeDataSetWrite: TXMLMapNotifyEvent read FBeforeDataSetWrite write FBeforeDataSetWrite;
		property OnRecordWrite: TXMLMapOnRecordWriteEvent read FOnRecordWrite write FOnRecordWrite;
		property BeforeRecordWrite: TXMLMapNotifyEvent read FBeforeRecordWrite write FBeforeRecordWrite;
		property OnFieldWrite: TXMLMapOnFieldWriteEvent read FOnFieldWrite write FOnFieldWrite;
		property AfterRecordWrite: TXMLMapNotifyEvent read FAfterRecordWrite write FAfterRecordWrite;
		property AfterDataSetWrite: TXMLMapNotifyEvent read FAfterDataSetWrite write FAfterDataSetWrite;

		property BeforeDataSetRead: TXMLMapNotifyEvent read FBeforeDataSetRead write FBeforeDataSetRead;
		property OnRecordRead: TXMLMapOnRecordReadEvent read FOnRecordRead write FOnRecordRead;
		property BeforeRecordRead: TXMLMapNotifyEvent read FBeforeRecordRead write FBeforeRecordRead;
		property OnFieldRead: TXMLMapOnFieldReadEvent read FOnFieldRead write FOnFieldRead;
		property AfterRecordRead: TXMLMapNotifyEvent read FAfterRecordRead write FAfterRecordRead;
		property AfterDataSetRead: TXMLMapNotifyEvent read FAfterDataSetRead write FAfterDataSetRead;
	end;

implementation

uses
	Controls, Dialogs, UTrace, XUtil;

{ TXMLMapField }

constructor TXMLMapField.Create(Collection: TCollection);
begin
	inherited Create(Collection);
	FDirection := xdReadWrite;
	FWriteCDATA := false;
	FField := nil;
end;

procedure TXMLMapField.SetFieldName(Value: string);
var
	d: TDataSet;
begin
	if FFieldName <> Value then
		begin
			if FNodeName = UpperCase(FFieldName) then
				FNodeName := UpperCase(Value);
			FFieldName := Value;
			d := TXMLMap(TXMLMapFields(Collection).GetOwner).DataSet;
			if d <> nil then
				FField := d.FindField(FFieldName)
			else
				FField := nil;
		end;
end;

function TXMLMapField.IsNodeNameStored: boolean;
begin
	result := FNodeName <> UpperCase(FFieldName);
end;

function TXMLMapField.IsFieldObjectStored: boolean;
var
	d: TDataSet;
begin
	d := TXMLMapFields(Collection).Owner.DataSet;
	result := (FField <> nil) and ((d = nil) or (FField <> d.FindField(FFieldName)));
end;

function TXMLMapField.GetDisplayName: string;
begin
	Result := FFieldName;
	if Result = '' then
		Result := inherited GetDisplayName;
end;

procedure TXMLMapField.Assign(Source: TPersistent);
var
	mf: TXMLMapField;
begin
	if Source is TXMLMapField then
		begin
			mf := TXMLMapField(Source);
			FFieldName := mf.FieldName;
			FNodeName := mf.NodeName;
			FWriteCDATA := mf.WriteCDATA;
		end
	else
		inherited Assign(Source);
end;

{ TXMLMapFields }

constructor TXMLMapFields.Create(AOwner: TXMLMap);
begin
	inherited Create(AOwner, TXMLMapField);
end;

function TXMLMapFields.GetItem(Index: Integer): TXMLMapField;
begin
	Result := TXMLMapField(inherited GetItem(Index));
end;

procedure TXMLMapFields.SetItem(Index: Integer; Value: TXMLMapField);
begin
	inherited SetItem(Index, Value);
end;

function TXMLMapFields.Add: TXMLMapField;
begin
	Result := TXMLMapField(inherited Add);
end;

function TXMLMapFields.Owner: TXMLMap;
begin
	result := TXMLMap(GetOwner);
end;

{ TXMLMap }

constructor TXMLMap.Create(AOwner: TComponent);
begin
	inherited;
	FMapFields := TXMLMapFields.Create(Self);
	FRecordNodeName := 'RECORD';
	FRecordsNodeName := 'RECORDS';
	FDateFormat := 'yyyymmdd';
	FTimeFormat := 'hhnnss';
	FDecimalSeparator := '.';
	FVersion := '1';
	FMaster := nil;
	FDetails := TList.Create;
end;

destructor TXMLMap.Destroy;
begin
	FMapFields.free;
	FDetails.Free;
	inherited;
end;

procedure TXMLMap.SetMapFields(Value: TXMLMapFields);
begin
	FMapFields.Assign(Value);
end;

procedure TXMLMap.SetDataSet(Value: TDataSet);
var
	i: integer;
begin
	if FDataSet <> Value then
		begin
			// First discard old field objects from previous dataset
			for i := 0 to FMapFields.Count-1 do
				if not MapFields[i].IsFieldObjectStored then
					MapFields[i].FieldObject := nil;
			// Assign new dataset
			FDataSet := Value;
			// Get field objects for new dataset
			if FDataSet <> nil then
				for i := 0 to FMapFields.Count-1 do
					if MapFields[i].FieldObject = nil then
						MapFields[i].FieldObject := FDataSet.FindField(MapFields[i].FieldName);
		end;
end;

procedure TXMLMap.SetMaster(Value: TXMLMap);
begin
	if FMaster <> Value then
		begin
			if FMaster <> nil then
				FMaster.UnRegisterDetail(Self);
			FMaster := Value;
			if FMaster <> nil then
				FMaster.RegisterDetail(Self);
		end;
end;

procedure TXMLMap.RegisterDetail(d: TXMLMap);
begin
	FDetails.Add(d);
end;

procedure TXMLMap.UnRegisterDetail(d: TXMLMap);
begin
	FDetails.Remove(d);
end;

procedure TXMLMap.DoBeforeDataSetWrite(Node: TXMLNode);
begin
	if Assigned(FBeforeDataSetWrite) then FBeforeDataSetWrite(Self, Node);
end;

procedure TXMLMap.DoBeforeRecordWrite(Node: TXMLNode);
begin
	if Assigned(FBeforeRecordWrite) then FBeforeRecordWrite(Self, Node);
end;

procedure TXMLMap.DoAfterRecordWrite(Node: TXMLNode);
begin
	if Assigned(FAfterRecordWrite) then FAfterRecordWrite(Self, Node);
end;

procedure TXMLMap.DoAfterDataSetWrite(Node: TXMLNode);
begin
	if Assigned(FAfterDataSetWrite) then FAfterDataSetWrite(Self, Node);
end;

procedure TXMLMap.DoOnRecordWrite(var Write: boolean);
begin
	if Assigned(FOnRecordWrite) then FOnRecordWrite(Self, Write);
end;

procedure TXMLMap.DoOnFieldWrite(MapField: TXMLMapField; var Write: boolean);
begin
	if Assigned(FOnFieldWrite) then FOnFieldWrite(Self, MapField, Write);
end;

procedure TXMLMap.DoBeforeDataSetRead(Node: TXMLNode);
begin
	if Assigned(FBeforeDataSetRead) then FBeforeDataSetRead(Self, Node);
end;

procedure TXMLMap.DoBeforeRecordRead(Node: TXMLNode);
begin
	if Assigned(FBeforeRecordRead) then FBeforeRecordRead(Self, Node);
end;

procedure TXMLMap.DoAfterRecordRead(Node: TXMLNode);
begin
	if Assigned(FAfterRecordRead) then FAfterRecordRead(Self, Node);
end;

procedure TXMLMap.DoAfterDataSetRead(Node: TXMLNode);
begin
	if Assigned(FAfterDataSetRead) then FAfterDataSetRead(Self, Node);
end;

procedure TXMLMap.DoOnRecordRead(Node: TXMLNode; var Action: TXMLMapRecordReadAction);
begin
	if Assigned(FOnRecordRead) then FOnRecordRead(Self, Node, Action);
end;

procedure TXMLMap.DoOnFieldRead(MapField: TXMLMapField; Node: TXMLNode; var Read: boolean);
begin
	if Assigned(FOnFieldRead) then FOnFieldRead(Self, MapField, Node, Read);
end;

function TXMLMap.WriteRecord: TXMLNode;

	function FieldToStr(f: TField): string;
	var
		SaveShortDateFormat: string;
		SaveLongTimeFormat: string;
		SaveDecimalSeparator: char;
	begin
		if f is TDateTimeField then
			begin
				SaveShortDateFormat := SysUtils.ShortDateFormat;
				SaveLongTimeFormat := SysUtils.LongTimeFormat;
				try
					SysUtils.ShortDateFormat := FDateFormat;
					SysUtils.LongTimeFormat := FTimeFormat;
					result := f.AsString;
				finally
					SysUtils.ShortDateFormat := SaveShortDateFormat;
					SysUtils.LongTimeFormat := SaveLongTimeFormat;
				end;
			end
		else if (f is TFloatField) or (f is TBCDField) then
			begin
				SaveDecimalSeparator := SysUtils.DecimalSeparator;
				try
					SysUtils.DecimalSeparator := FDecimalSeparator;
					result := f.AsString;
				finally
					SysUtils.DecimalSeparator := SaveDecimalSeparator;
				end;
			end
		else if f is TBooleanField then
			result := IntToStr(Ord(f.AsBoolean))
		else
			result := f.AsString;
	end;

var
	i: integer;
	mf: TXMLMapField;
	Write: boolean;
begin
	result := TXMLNode.Create;
	try
		result.NodeName := FRecordNodeName;
		DoBeforeRecordWrite(result);
		for i := 0 to FMapFields.Count-1 do
			begin
				mf := FMapFields[i];
				if (mf.Direction in [xdWriteOnly, xdReadWrite]) and (mf.FieldObject <> nil) and (not mf.FieldObject.IsNull) then
					begin
						Write := true;
						DoOnFieldWrite(mf, Write);
						if Write then
							begin
								if mf.WriteCDATA then
									result.Elements[mf.NodeName].CreateChild(xntCData).NodeValue := FieldToStr(mf.FieldObject)
								else
									result.Elements[mf.NodeName].CreateChild(xntText).NodeValue := FieldToStr(mf.FieldObject);
							end;
					end;
			end;
		for i := 0 to FDetails.Count-1 do
			result.AppendChild(TXMLMap(FDetails[i]).WriteRecords);
		DoAfterRecordWrite(result);
	except
		result.Free;
		raise;
	end;
end;

function TXMLMap.WriteRecords: TXMLNode;
var
	Count: integer;
	Write: boolean;
begin
	result := TXMLNode.Create;
	try
		result.NodeName := FRecordsNodeName;
		if DataSet <> nil then
			begin
				DoBeforeDataSetWrite(result);
				DataSet.First;
				Count := 0;
				while not DataSet.Eof do
					begin
						Write := true;
						DoOnRecordWrite(Write);
						if Write then
							begin
								result.AppendChild(WriteRecord);
                Inc(Count);
              end;
						DataSet.Next;
					end;
				result.Attributes['COUNT'].NodeValue := IntToStr(Count);
				result.Attributes['VERSION'].NodeValue := FVersion;
				DoAfterDataSetWrite(result);
			end;
	except
		result.Free;
		raise;
	end;
end;

procedure TXMLMap.ReadRecord (Node: TXMLNode);

	procedure StrToField(s: string; f: TField);
	var
		pos: integer;
		SaveDecimalSeparator: char;
		float_value: double;
	begin
		if f is TDateField then
			begin
				pos := 1;
				f.AsDateTime := XScanDateTime(s, FDateFormat, pos);
			end
		else if f is TTimeField then
			begin
				pos := 1;
				f.AsDateTime := XScanDateTime(s, FTimeFormat, pos);
			end
		else if f is TDateTimeField then
			begin
				pos := 1;
				f.AsDateTime := XScanDateTime(s, FDateFormat, pos) + XScanDateTime(s, FTimeFormat, pos);
			end
		else if (f is TFloatField) or (f is TBCDField) then
			begin
				SaveDecimalSeparator := SysUtils.DecimalSeparator;
				try
					SysUtils.DecimalSeparator := FDecimalSeparator;
					float_value := StrToFloat(s);
				finally
					SysUtils.DecimalSeparator := SaveDecimalSeparator;
				end;
				f.AsFloat := float_value;
			end
		else if f is TBooleanField then
			f.AsBoolean := Boolean(StrToInt(s))
		else
			f.AsString := s;
	end;

var
	i: integer;
	mf: TXMLMapField;
	n: TXMLNode;
	Read: boolean;
begin
	DoBeforeRecordRead(Node);
	for i := 0 to FMapFields.Count-1 do
		begin
			mf := FMapFields[i];
			if (mf.Direction in [xdReadOnly, xdReadWrite]) and (mf.FieldObject <> nil) then
				begin
					n := Node.FindChild(mf.NodeName);
					Read := true;
					DoOnFieldRead(mf, n, Read);
					if Read then
						begin
							if n <> nil then
								StrToField(n.Data, mf.FieldObject)
							else
								mf.FieldObject.Clear;
						end;
				end;
		end;
	for i := 0 to FDetails.Count-1 do
		TXMLMap(FDetails[i]).ReadRecords(Node.Elements[TXMLMap(FDetails[i]).RecordsNodeName]);
	DoAfterRecordRead(Node);
end;

procedure TXMLMap.ReadRecords (Node: TXMLNode);
var
	r: TXMLNode;
	Action: TXMLMapRecordReadAction;
begin
	if (DataSet <> nil) and (Node <> nil) then
		begin
			DoBeforeDataSetRead(Node);
			r := Node.FirstChild;
			while r <> nil do
				begin
					if (r.NodeType = xntElement) and (r.NodeName = FRecordNodeName) then
						begin
							Action := xnaAppend;
							DoOnRecordRead(r, Action);
              if Action <> xnaSkip then
                begin
                  if Action = xnaAppend then
                    DataSet.Append
                  else
                    DataSet.Edit;
                  ReadRecord(r);
                  DataSet.Post;
                end;
						end;
					r := r.NextSibling;
				end;
			DoAfterDataSetRead(Node);
		end;
end;

end.




