{: XMLNode

PURPOSE AND IMPLEMENTATION:
	This unit defines classes for parsing and writing a well-formed standalone
	XML document in	to a tree of nodes describing the data therein.
	These classes only support a subset of the XML 1.0 recommendation and have
	not been analyzed for conformance with the recommendation.
	This module also only supports 8-bit character sets and makes no special
	allowances for UTF-8. Nor has any support for the "<!DOCTYPE" node been
	implemented. Parsing of the XML declaration lazily uses the attribute parsing
	code which is not correct (but works).

	Q:  Why use this module instead of Microsoft's XML parser or the OpenXML
			project?

	A:  I wrote this module to be free of dependencies on outside libraries and to
			be small and fast.  Neither of the above libraries met that criteria.  I
			didn't like the idea of requiring my users to install Microsoft's XML
			library (and the potential compatibility problems that arise).  Nor did I
			like the idea of adding 135 kb (OpenXML) to my executable just to load files
			in XML format.

	This should work for Delphi 4-7.

HISTORY:
	09/12/00 - Created by Colin Patrick Sarsfield

DISCLAIMER:
	Use at your own risk.

COPYRIGHT:
	Released into the public domain by Colin Sarsfield 9/13/00
}

{: XMLNode

	Original XMLReader and XMLWriter from Colin repackaged by Tomislav Karda (tomislav.kardas@pmx.hr)
	into one XMLNode unit and a lot of changes/improvements made to the original code.

	- TMXLNode class improved
			- parent, data, text, exacttext, xml properties added
			- Attributes and Elements	array properties added
			- TXMLNodeType extended for two new node types: xntText, xntCData to support Mixed Content
			- some things removed, like saving node data AsFloat, AsDateTime,
				etc. because it depends on system settings. Only string access left.
			- tried to follow names that MSXML uses for properties and method of the node object
			- added simple node selection with selectNodes and selectSingleNode functions that
				is a very basic XPath locator

	- improved XML readers
			- from string, stream and file reader
			- events added to XMLReader so that parsing can be intercepted when node completes, etc.
			- doctype parsing added just to store the info in the node, not to abort parsing of such documents
			- PreserveWhiteSpace, MixedContent, ProgressPercent, Size, Position - properties added to reader
			- Text and CData node types added, and Mixed Content parsing supported. xntText & xntCData nodes
				implemented in the ParseContent	procedure where for each text or CData section new node	is created
				instead of accumulating	all text into element node.
			- LineNumber, ColNumber property added
			- HTMLReader property added to reader, to support parsing of HTML documents.
				Since HTML is similar to XML, I thought it would be a great thing to be able to load HTML as a node tree.
				So, if you have to extract some information from a HTML file that you can do through tree structure.
				This is done on a flag, so regular XML parsing is not affected. I have done it on the try and correct basis,
				and I can tell you HTML is such a loose format. I hope I covered a majority of cases, it may happen
				that some HTML pages wouldn't load.

	- improved XML writers
			- to string, stream and file writer
			- events added to XMLWriter so that writing can be intercepted on certain events.
			- FormatOutput, ProgressPercent, Position, Total, Count - properties added to writer
			- WriteNode procedure Added to XMLWriter for full node write in one method call

	- XML reader & writer fully optimized for Speed and Memory requirement. Now it outperforms
		msxml when reading huge files.
			- Reading of node elements is buffered through PChar buffers. This option is added because
				otherwise strings take a lot of virtual memory and specialy when large files are parsed (few MB),
				it also increases the parsing speed rapidly. Virtual memory requirements dramaticaly reduced
				since there is no string handling around.
			- child nodes are implemented without TList object, using double linked lists, this increase
				memory usage a little, but gains a lot on speed because new nodes are inserted at the end of the list
				by just changing a few pointers. There is no need to maintain separate TList object to hold pointers
				to every child node. This also gains in speed since there is no need to realocate TList memory when number
				of child nodes increases. But this also imposes a different programming style since there is no index
				approach to child nodes, rather you can approach to first and last child nodes and then skip forward
				and backward. I prefere this way since it's faster and more natural.
			- Previous system of SaveState and RestoreState technique while parsing is changed to unread buffer.
				Chars are pulled out from the xml source and if parser needs to go back it returns
				chars back (unreads them) and they are stored into the unread buffer. Subsequent reads first check
				the unread buffer and then if empty pulls the xml source.

	- TXMLDoc component added for easy using of XMLNode, XMLWriter and XMLReader objects through
		simple methods for loading and saving from/to different sources/destination and using
		events to accomplish progress counter or alter the process of load/save.

	- Demo application improved
}

unit XMLNode;

interface

uses
	Classes, SysUtils;

{$define PARENT1}

type
	TXBuf = object
	private
		FSize: integer;
		FPtr: PChar;
		function GetAsString: string;
		procedure SetAsString(value: string);
		procedure SetSize(value: integer);
	public
		procedure Init;
		procedure Clear;
		procedure CopyFrom(SourcePtr: PChar; SourceSize: integer);
		procedure AppendFrom(SourcePtr: PChar; SourceSize: integer);
		property AsString: string read GetAsString write SetAsString;
		property Size: integer read FSize write SetSize;
		property Ptr: PChar read FPtr;
	end;

type
	CharSet = set of Char;

	TXMLNodeType = (
		xntDocument,
		xntXMLDeclaration,
		xntDocumentType,
		xntElement,
		xntAttribute,
		xntText,
		xntCData,
		xntProcessingInstruction
	);

	(*
		There is no programmed limitation to what node types can be childs to other node types,
		but users should hold to following logic which is supported by Parser and Writter:

		xntDocument:
			xntXMLDeclaration
			xntDocumentType
			xntProcessingInstruction
		xntXMLDeclaration:
			xntAttribute
		xntElement:
			xntAttribute
			xntText
			xntCData
			xntElement
			xntProcessingInstruction
	*)

const
	TXMLNodeTypeStrings: array[TXMLNodeType] of String = (
		'Document',
		'XMLDeclaration',
		'DocumentType',
		'Element',
		'Attribute',
		'Text',
		'CDataSection',
		'ProcessingInstruction'
	);

type
	TXMLNode = class;
	TXMLNodeArray = array of TXMLNode;
	TXMLNode = class
	private
		FParentNode: TXMLNode;
		FNodeType: TXMLNodeType;

		FNodeName: String;
		FNodeValue: String;

		FPreviousSibling: TXMLNode;			// Previous sibling
		FNextSibling: TXMLNode;			// Next sibling
		FChildNodesCount: integer;	// Number of child nodes
		FFirstChild: TXMLNode;	// First child node
		FLastChild: TXMLNode;	// Last child node

		procedure SetParent(AParent: TXMLNode);
		function GetOwnerDocument: TXMLNode;
		function GetPath: string;
		function GetTotalNodesCount: Integer;
		function GetPreviousNode: TXMLNode;
		function GetNextNode: TXMLNode;

		function GetAttributes(Name: string): TXMLNode;
		function GetElements(Name: string): TXMLNode;

		function GetData: string;
		procedure SetData(value: string);
		function GetText: string;
		function GetExactText: string;
		function GetXml: string;
	public
		// default NodeType is xntElement, Name and Data are ''
		constructor Create(const NodeType: TXMLNodeType = xntElement);
		// when destroyed, node will destroy all it's child nodes
		destructor Destroy; override;

		property NodeName: String read FNodeName write FNodeName;
		property NodeValue: String read FNodeValue write FNodeValue;
		property NodeType: TXMLNodeType read FNodeType write FNodeType;
		function Description: String;

		// ParentNode:
		//      	Represents node that is a parent to the node. ParentNode property is automaticaly
		//				set by using NewChild & AppendChild methods, and cleared when using RemoveChild
		//				on the parent node.
		//				If assigned directly it will remove the node from previous parent and add it as a
		//				child to new parent if new parent is not nil.
		//	OwnerDocument:
		//				Document that contains this node. It's a root node that has no parent.
		//	Path:
		//				Path to this node made of NodeNames of parent nodes starting from the root node
		property ParentNode: TXMLNode read FParentNode write SetParent;
		property OwnerDocument: TXMLNode read GetOwnerDocument;
		property Path: String read GetPath;

		// FirstChild:
		//      	First child node object to this node.
		// LastChild:
		//      	Last child node object to this node.
		// PreviousSibling:
		//      	Previous sibling node object to this node. This is a node that has same parent as this node but
		//				is right before him in the list of child nodes for it's parent.
		// NextSibling:
		//      	Next sibling node object to this node. This is a node that has same parent as this node but
		//				is right after him in the list of child nodes for it's parent.
		// PreviousNode:
		//      	Previous node object to this node. This is a PreviousSibling's last child if there are one,
		//				or is a PreviousSibling if node is not the firt child of it's	parent, otherwise it's a parent node.
		// NextNode:
		//      	Next node object to this node. This is a FirstChild if there are childs, or NextSibling if node is
		//				not the last child of it's parent, otherwise it's a parent's NextSibling node, etc.
		// ChildNodesCount:
		//      	Number of child nodes for this node.
		//
		// GetFirstChildNode:
		//        Gets first child node of given type from current node acting as parent.
		// GetNextSiblingNode:
		//        Gets next sibling node of given type from current node.
		property FirstChild: TXMLNode read FFirstChild;
		property LastChild: TXMLNode read FLastChild;
		property PreviousSibling: TXMLNode read FPreviousSibling;
		property NextSibling: TXMLNode read FNextSibling;
		property PreviousNode: TXMLNode read GetPreviousNode;
		property NextNode: TXMLNode read GetNextNode;
		property ChildNodesCount: Integer read FChildNodesCount;

		function GetFirstChildNode(const NodeType: TXMLNodeType = xntElement): TXMLNode;
		function GetNextSiblingNode(const NodeType: TXMLNodeType = xntElement): TXMLNode;

		// TotalNodesCount:
		// 				Total Node Count, including all child nodes in the three below this node
		property TotalNodesCount: Integer read GetTotalNodesCount;

		// CreateChild:
		//				Creates new child node object
		function CreateChild(const NodeType: TXMLNodeType = xntElement): TXMLNode;
		// AppendChild:
		//				Adds existing node object as a child for current node,
		// 				if node is a child of other parent, reference is automaticaly removed
		// RemoveChild:
		//				removes node from the list of child nodes,
		//				node's parent is set to nil, and node is not destroyed.
		//				returns true if node was a child node.
		//				Note: To destroy a child node simple call it's free method,
		//				this will remove it from the child list and destroy
		procedure AppendChild(Node: TXMLNode);
		function RemoveChild(Node: TXMLNode): boolean;

		// CloneNode:
		//			 		Clones node and makes a new copy of it. The clone has no parent node and it is a duty of
		//					application to append it as client to some node or to free it manualy when not needed.
		//					If deep is set then all the subtree is cloned together with this node.
		function CloneNode(deep: boolean): TXMLNode;

		function HasChildNodes: boolean;

		// FindChild:
		//					finds n'th (Index) occurence of a node with given Name and NodeType.
		//					if node not found returns nil, or if CreateMissing specified then
		//					new node is created and returned as found.
		function FindChild(const NodeName: String; const NodeType: TXMLNodeType = xntElement;
			const Index: Integer = 0; const CreateMissing: boolean = false): TXMLNode;

		// Attributes:	access to node of type xntAttribute via node name
		// Elements:		access to node of type xntElement via node name
		property Attributes[Name: string]: TXMLNode read GetAttributes;
		property Elements[Name: string]: TXMLNode read GetElements;

		// Data:
		//					When read, Data property gives NodeValue plus all NodeValues of child Text and CData nodes.
		//					When write, Data property stored NodeValue and destroys all child Text and CData nodes.
		//					Data property gives a flexible and easier approach to simple elements and eliminates the trouble
		//					to read element node values when document is loaded with MixedContent = true. Using this Data
		//					property application doesn't have to worry if element value is actualy stored in the element node
		//					itself or in it's child Text and CData nodes. It's whole purpose is for DataSet fields mapping to
		//					xml element nodes to simplify processing of reading and writing field values from/to element nodes.
		// Text:
		//					Gives text representation of the node. Text is built from this node and all the child Element,
		//					Text and CData nodes by adding theirs NodeValues. The resulting text is trimmed of white space
		//					from the beggining and end. Special handling is added to detect situations where two consequtive
		// 					non empty element nodes are combined in one text string to separate them with space.
		// 					This might not be the best approach, since when white space is significant (like in html)
		//					this adds extra space where it should not.
		//					The example of unwanted behaviour is: '<NODE><U>First</U><I><B>Second</B></I></NODE>'
		//					Where Text would be: 'First Second' while actually 'FirstSecond' should be returned.
		//					For such application use ExactText property.
		// ExactText:
		//					The same as Text, but no extra space added between two consequtive non empty element nodes.
		// Xml:
		//					gives the xml string of the subtree

		property Data: String read GetData write SetData;
		property Text: String read GetText;
		property ExactText: String read GetExactText;
		property Xml: String read GetXml;

		(*
				selectNodes:
								Gives a list of nodes that matches a basic XPath location path when applied
								from the current node.
				selectSingleNode:
								Same as above, but returns only the first node selected.

				queryString is a Location Path specification, supported are:
				-	Absolute & relative location path. Absolute location path starts with /
				- Multiple location steps, every step is separated with /
				- Abbreviated syntax only
				- Supported axes: parent, self, attribute, child, descendant
				- Node test is only name test (element or attribute name with * as a wildchar)
				- Only one predicate supported, only index integer predicate without any functions.

				Examples that work:
				'para' 		selects the para element children of the context node
				'*' 			selects all element children of the context node
				'@name' 	selects the name attribute of the context node
				'@*' 			selects all the attributes of the context node
				'para[1]'	selects the first para child of the context node
				'*/para'	selects all para grandchildren of the context node
				'/doc/chapter[5]/section[2]'	selects the second section of the fifth chapter of the doc
				'chapter//para'								selects the para element descendants of the chapter element
																			children of the context node
				'//para'		selects all the para descendants of the document root and thus selects all para
										elements in the same document as the context node
				'//olist/item'		selects all the item elements in the same document as the context node that
													have an olist parent
				'.'					selects the context node
				'.//para'		selects the para element descendants of the context node
				'..'				selects the parent of the context node
				'../@lang'	selects the lang attribute of the parent of the context node

				Future extensions planned:
				- Predicates of form [elemname], [elemname='value'], [attrname], [attrname='value'] for simple
					selections based on the values of attributes or child elements.
				- Support for multiple predicates, so that index can be applied after the above selections,
					or and like functionality for existence of elements and attriutes.
		*)
		function selectNodes(const queryString: String): TXMLNodeArray;
		function selectSingleNode(const queryString: String): TXMLNode;

		(*
		function getElementsByTagName(const tagName: String): TXMLNodeList;
		*)
	end;

(*
		Node utility functions:

		NodeNameMatch
					compares given node name with a match template, case sensitive, * is a jocker char
		AddNodeToArray
					expands nodes array and adds node element at the end
		AddNodesToArray
					expands nodes array and adds node elements from the node array at the end
		PackNodeArray
					deletes all nil pointers from node array
		IsNodeInArray
					checkes if the given not is part of the array
*)

function NodeNameMatch(cName, cMatch: string): boolean;
procedure AddNodeToArray(var aNodes: TXMLNodeArray; Node: TXMLNode);
procedure AddNodesToArray(var aNodes: TXMLNodeArray; AddNodes: TXMLNodeArray);
procedure PackNodeArray(var aNodes: TXMLNodeArray);
function IsNodeInArray(var aNodes: TXMLNodeArray; Node: TXMLNode): boolean;

const
(*
	XMLStreamReaderCacheSize -  Cache size (64 KB)
	XMLStreamReaderCacheSave -  Min. size of behind cache (256 B)
	XMLStreamReaderLoadMask  -  Mask for rounding data size loading from stream, round to 1KB
*)
	XMLStreamReaderCacheSize  = $010000;

type
	TXMLReader = class
	private
		CurChar: Char;

		FPreserveWhiteSpace: boolean;
		FMixedContent: boolean;
		FHTMLReader: boolean;
		FProgressStep: integer;

		(*
			FProgressCount:
				Number of bytes that has to be read before calling each OnProgress event
			FProgressPos:
				Current countdown of read bytes before calling next OnProgress event
		*)
		FProgressCount: integer;
		FProgressPos: integer;
		FProgressPercent: integer;
		FLineNumber, FColNumber: integer;
		FLineState: integer;	// interni state brojaa linija
		FNode: TXMLNode;

		FOnNode: TNotifyEvent;
		FOnElementStart: TNotifyEvent;
		FOnElementEnd: TNotifyEvent;
		FOnProgress: TNotifyEvent;

	private
		(*
			Read buffer intruduced here to avoid using string type in character data processing
			since string + char operation takes much more time than storing byte in a buffer

			ReadBufAdd:
							adds one char to read buff.
			ReadBufClose:
							closes read buffer from the given position and returns buffered string that has accumulated
							from that position.
		*)
		FReadBuf: TXBuf;
		FReadBufPos: integer;
		procedure ReadBufAdd(c: char);
		function ReadBufClose(p: integer): TXBuf;

	private
		(*
			UnRead buffer intruduced here to avoid requesting xml source to go backwards when needed. Temporary
			chars that have been read will be placed here for next char reading.

			UnReadChar:
							Unreads one character.
			UnReadString:
							Unreads a character sequence. Note here that character sequences are unread in reverse order
							(from last to first char) so that when reading again it will read the same string. Unreading
							is a LIFO (last in first out) structure.
		*)
		FUnReadBuf: TXBuf;
		FUnReadBufPos: integer;
		procedure UnReadChar(c: char);
		procedure UnReadString(p: PChar; count: integer);

		function DoesStreamMatchString(Str: String; CaseSensitive: boolean = true): Boolean;
		procedure ReadChar;
		procedure ReadCharAndCheckForEnd;
		procedure ParseAttribute(Node: TXMLNode);
		procedure ParseComment;
		procedure ParseContent(Node: TXMLNode; EndTagName: string);
		procedure ParseContentAsIs(var Content: TXBuf; EndSequence: string; CaseSensitive: boolean = true);
		procedure ParseElement(Node: TXMLNode);
		function ParseEq: Boolean;
		procedure ParseName(var Name: TXBuf);
		procedure ParseProcessingInstruction(Node: TXMLNode);
		function ParseReference: Char;
		function ParseWhitespace: Boolean;
		procedure ReportError(Err: String);
	protected
		procedure DoNode(Node: TXMLNode);
		procedure DoElementStart(Node: TXMLNode);
		procedure DoElementEnd(Node: TXMLNode);
		procedure DoProgress;

		function GetSize: integer; virtual; abstract;
		function GetPosition: integer; virtual; abstract;
		function EndOfData: Boolean; virtual; abstract;
		// GetChar returns next character in the input source, or #0 if end of source
		function GetChar: char; virtual; abstract;
	public
		constructor Create;
		destructor Destroy; override;

		// PreserveWhiteSpace:
		// 				If set then all the white space that appears between elements will be parsed
		//				into the document. Documents that were written with FormatOutput = true, should be read with
		// 				PreserveWhiteSpace set to false so that whitespace used for output formating would not be parsed
		//				in and waist memory and create unnecesary text objects.
		//				This property can be changed in the reader events and in that way alter the parsing of different nodes.
		// MixedContent:
		// 				If not set then all the text and CData inside the node goes directly to NodeValue of the element.
		//				Otherwise text and CData are stored in separate child nodes of the element node. If XML document
		//				that is beeing read is known to not have mixed content then this property can be turned off
		//				to not create separate node for every text and CData. There is never a problem if you always leave
		//				MixedContent = true and use Data property of the element node to read simple elements.
		//				This property can be changed in the reader events and in that way alter the parsing of different nodes.
		// HTMLReader:
		// 				If set then:
		//				1.	HTML elements that have no endings tag will be parsed in without error raised.
		//						HTML elements that will be read this way are: META, IMG, LINK, HR, BR, INPUT
		//				2.	Argument values are being parsed if without quotes since HTML pages frequently
		//						have arguments this way. Also arguments doesn't have to have values. If only
		//						argument name is present, it is added as blank value argument.
		//				3.  &nbsp; is being parsed as space.
		//				4.	Tag names are treated case insensitive when matching end and start tags of a node.
		//				5.	Start tags can be left unclosed: <tag attr=value<tag2> is parsed as <tag attr=value><tag2>
		//						If new node starts and previous tag was not closed, it is closed automaticaly.
		//				6.  <SCRIPT> node is loaded as is, without any parsing - up to </SCRIPT> tag.
		//				7.	when close tag is found that does not match current open tag - then: if there is any node
		//						in parent path with that tag name still open - close all nodes up to him.
		//						Otherwise just ignore this offending close tag.
		// ProgressStep:
		//				Percent step value, how many percents of job done have to change to call OnProgress event,
		//				default is 5 what means that OnProgress event is called on 0, 5, 10, 15, ... 100 percent of progress.
		//				If this property need to be changed, it only takes effect if set before calling of ParseDocument, do
		//				not change it in reader events, since that would make ProgressPercent property out of sense.
		property PreserveWhiteSpace: boolean read FPreserveWhiteSpace write FPreserveWhiteSpace default false;
		property MixedContent: boolean read FMixedContent write FMixedContent default false;
		property HTMLReader: boolean read FHTMLReader write FHTMLReader default false;
		property ProgressStep: integer read FProgressStep write FProgressStep default 5;

		// ParseDocument:
		// 				Fills the XML document structure from the XMLReader source
		procedure ParseDocument(Document: TXMLNode);

		// Size:
		// 				Total size in bytes of XML text in XMLReader source
		// Position:
		// 				Current reader position in XML text in XMLReader source
		// ProgressPercent:
		// 				Current percent of read job done, from 0-100 in step of 5
		// LineNumber, ColNumber:
		//				Line and column numbers where the parser currently reads
		// Node:
		// 				Current Node when OnNode, OnElementStart, OnElementEnd executes
		property Size: integer read GetSize;
		property Position: integer read GetPosition;
		property ProgressPercent: integer read FProgressPercent;
		property LineNumber: integer read FLineNumber;
		property ColNumber: integer read FColNumber;
		property Node: TXMLNode read FNode;

		property OnNode: TNotifyEvent read FOnNode write FOnNode;
		property OnElementStart: TNotifyEvent read FOnElementStart write FOnElementStart;
		property OnElementEnd: TNotifyEvent read FOnElementEnd write FOnElementEnd;
		property OnProgress: TNotifyEvent read FOnProgress write FOnProgress;
	end;

	TXMLStringReader = class(TXMLReader)
	private
		FData: String;
		FPosition: Integer;
	protected
		function GetSize: integer; override;
		function GetPosition: integer; override;
		function EndOfData: Boolean; override;
		function GetChar: char; override;
	public
		constructor Create(AData: string);
		property Data: String read FData;
	end;

	TXMLStreamReader = class(TXMLReader)
	private
		FStream: TStream;
		FSize: integer;			// Stream size
		FPosition: integer;	// Current stream position

		FCacheBuf: array [0..XMLStreamReaderCacheSize-1] of char; // Stream cache
		FCacheLen: integer; // How many chars is in the cache

		// How many chars we are behind the current stream position inside the cache,
		// if 0 then we have exhausted the cache and have to read from stream, if not read from cache.
		FBehind: integer;
	protected
		function GetSize: integer; override;
		function GetPosition: integer; override;
		function EndOfData: Boolean; override;
		function GetChar: char; override;
	public
		constructor Create(AStream: TStream);
		property Stream: TStream read FStream;
	end;

	TXMLFileReader = class(TXMLStreamReader)
	private
		FFileName: string;
	protected
	public
		constructor Create(AFileName: string);
		destructor Destroy; override;
		property FileName: string read FFileName;
	end;

const
	XMLWriterCacheSize = $010000;

type
	TXMLTagType = (xttStarting, xttStartingAndEnding, xttEnding, xttEndingSameLine);

	TXMLWriter = class
	private
		FWriteBuf: TXBuf;
		FWriteBufPos: integer;
		FIndentation: Integer;
		FFirstLine: boolean;

		FFormatOutput: boolean;
		//FProgressStep: integer;

		FPosition: integer;
		FTotal: integer;
		FCount: integer;
		FProgressPercent: integer;
		FNode: TXMLNode;

		FOnElementStart: TNotifyEvent;
		FOnElementEnd: TNotifyEvent;
		FOnProgress: TNotifyEvent;
	protected
		procedure WriteBuffer(Buffer: PChar; Count: integer); virtual;
		procedure WriteChar(c: char);

		procedure DoElementStart(Node: TXMLNode);
		procedure DoElementEnd(Node: TXMLNode);
		procedure DoProgress;
	public
		constructor Create;
		destructor Destroy; override;

		// FormatOutput
		//				If set then white space is added so that document output looks nice when viewed/edited
		//				as text file. This can be done only if white space is not important in document content
		//				and readings should be parsed in XMLReader with PreserveWhiteSpace = false.
		property FormatOutput: boolean read FFormatOutput write FFormatOutput default true;

		// ProgressStep:
		//				Percent step value, how many percents of job done have to change to call OnProgress event,
		//				default is 5 what means that OnProgress event is called on 0, 5, 10, 15, ... 100 percent of progress.
		//				If this property need to be changed, it only takes effect if set before calling of WriteNode, do
		//				not change it in writer events, since that would make ProgressPercent property out of sense.
		//property ProgressStep: integer read FProgressStep write FProgressStep default 5;

		// Position
		//				Number of characters already written to the Writer destination. There is no Size property
		//				since Size is not known until end, but Total & Count are used for that purpose since
		//				that gives an idea of where we are right now in the process of writing.
		// Total
		//				Total number of nodes that has to be written to the destination. It is set in WriteNode
		//				method when invoked after StartDoc.
		// Count
		//				Number of nodes that are already written to the destination.
		// ProgressPercent
		//				Percent of write job done. It is calculated as: (100 * Count) div Total.
		// Node:
		// 				Current Node when OnElementStart, OnElementEnd executes
		property Position: integer read FPosition;
		property Total: integer read FTotal;
		property Count: integer read FCount;
		property ProgressPercent: integer read FProgressPercent;
		property Node: TXMLNode read FNode;

		// StartDoc
		//				Prepares writter for writing.
		// WriteNode
		//				Writes complete node with all child elements to destination.
		// EndDoc
		//				Finishes writting.
		//
		// Note:	The easiest way for application to write XML document is to construct TXMLNode object that
		//				represents whole Document, fill it with appropriate child elements and then call:
		//				StartDoc, WriteNode, EndDoc. Other write methods below are automaticaly called by WriteNode
		//				but can be used directly if application wants to write without constructed TXMLNode document
		//				object or if only parts of XML document are constructed in TXMLNode objects. WriteNode can be
		//				mixed with other Write methods.
		procedure StartDoc;
		procedure WriteNode(Node: TXMLNode);
		procedure EndDoc;

		procedure WriteStr(const Value: String);
		procedure Write(const Value: String);
		procedure WriteCDATA(Data: string);
		procedure WriteData(const Name: String; const Data: String);
		procedure WriteStandaloneDocumentDeclaration;
		procedure WriteTag(const Name: String; TagType: TXMLTagType = xttStarting; Open: Boolean = False);
		procedure WriteTagParam(const ParamName, Value: String);
		procedure WriteTagClose(TagType: TXMLTagType = xttStarting);
		// WriteLineEnd does nothing if FormatOutput = false, otherwise outputs newline (CRLF) and idents
		// with spaces according to the open tag nesting level.
		procedure WriteLineEnd;

		property OnElementStart: TNotifyEvent read FOnElementStart write FOnElementStart;
		property OnElementEnd: TNotifyEvent read FOnElementEnd write FOnElementEnd;
		property OnProgress: TNotifyEvent read FOnProgress write FOnProgress;
	end;

	TXMLStringWriter = class(TXMLWriter)
	private
		FString: PString;
	protected
		procedure WriteBuffer(Buffer: PChar; Count: integer); override;
	public
		constructor Create(var AString: string);
	end;

	TXMLStreamWriter = class(TXMLWriter)
	private
		FStream: TStream;
	protected
		procedure WriteBuffer(Buffer: PChar; Count: integer); override;
	public
		constructor Create(AStream: TStream);
		property Stream: TStream read FStream;
	end;

	TXMLFileWriter = class(TXMLStreamWriter)
	private
		FFileName: string;
	public
		constructor Create(AFileName: string);
		destructor Destroy; override;
		property FileName: string read FFileName;
	end;

type
	TXMLReaderEvent = procedure (Sender: TObject; Reader: TXMLReader) of object;
	TXMLWriterEvent = procedure (Sender: TObject; Writer: TXMLWriter) of object;

type
	TXMLDoc = class(TComponent)
	private
		FDocument: TXMLNode;
		FIsEmpty: boolean;
		FLoading: boolean;
		FSaving: boolean;

		FPreserveWhiteSpace: boolean;
		FMixedContent: boolean;
		FHTMLReader: boolean;
		FProgressStep: integer;
		FFormatOutput: boolean;

		FOnReadDocumentStart: TXMLReaderEvent;
		FOnReadDocumentEnd: TXMLReaderEvent;
		FOnReadNode: TXMLReaderEvent;
		FOnReadElementStart: TXMLReaderEvent;
		FOnReadElementEnd: TXMLReaderEvent;
		FOnReadProgress: TXMLReaderEvent;

		FOnWriteDocumentStart: TXMLWriterEvent;
		FOnWriteDocumentEnd: TXMLWriterEvent;
		FOnWriteElementStart: TXMLWriterEvent;
		FOnWriteElementEnd: TXMLWriterEvent;
		FOnWriteProgress: TXMLWriterEvent;

		procedure	DoReadDocumentStart(Reader: TXMLReader);
		procedure	DoReadDocumentEnd(Reader: TXMLReader);
		procedure	DoReadNode(Sender: TObject);
		procedure	DoReadElementStart(Sender: TObject);
		procedure	DoReadElementEnd(Sender: TObject);
		procedure	DoReadProgress(Sender: TObject);

		procedure	DoWriteDocumentStart(Writer: TXMLWriter);
		procedure	DoWriteDocumentEnd(Writer: TXMLWriter);
		procedure	DoWriteElementStart(Sender: TObject);
		procedure	DoWriteElementEnd(Sender: TObject);
		procedure	DoWriteProgress(Sender: TObject);

		procedure CheckBusy;
		procedure LoadFromReader(r: TXMLReader);
		procedure SaveToWriter(w: TXMLWriter);
	public
		constructor Create(AOwner: TComponent); override;
		destructor Destroy; override;

		// Document:
		// 				Document node that contains all the nodes, including XML declaration, root document element, etc...
		//				this is a complete XML file in tree object model that can either be filled out programaticaly
		//				using	CreateNode method or loaded from some XML source.
		property Document: TXMLNode read FDocument;

		// Clear:
		// 				Frees any nodes from Document node. Complete clear, as was after creating a component.
    //        If CreateXmlDecl is set then following structure will be automaticaly created in the document:
    //        <?xml version='1.0' encoding='windows-1250'?>
		procedure Clear(CreateXmlDecl: boolean = true);

		(*
		property DocumentElement: TXMLNode read GetDocumentElement write SetDocumentElement;
		function createNode(NodeType: TXMLNodeType; const Name: String): TXMLNode;
		function createElement(const Name: String): TXMLNode;
		function createAttribute(const Name: String): TXMLNode;
		*)

		// LoadFromXXX:
		//        Loads a complete XML Document into Document object model into Document property,
		//				any previous Document object tree is removed and new object model from source is placed.
		procedure LoadFromString(xmlstring: string);
		procedure LoadFromStream(xmlstream: TStream);
		procedure LoadFromFile(xmlfilename: string);

		// SaveToXXX:
		//        Saves complete XML Document in Document object model from Document property into
		//				destination media.
		function SaveToString: string;
		procedure SaveToStream(xmlstream: TStream);
		procedure SaveToFile(xmlfilename: string);

		// IsEmpty:
		//        Indicates whether the document is Empty. It is automaticaly set after document Load, and
		//				cleared after document Clear. In addition to that, application can manualy alter this property
		//				when document is filled out in application, but there is no need to do so.
		// Loading:
		//				If true, means that loading of xml document is in progress
		// Saving:
		//				If true, means that saving of xml document is in progress
		property IsEmpty: boolean read FIsEmpty write FIsEmpty;
		property Loading: boolean read FLoading;
		property Saving: boolean read FSaving;
	published

		// PreserveWhiteSpace, MixedContent, HTMLReader:
		//				See help for this properties on TXMLReader object. This values here serve as a starting
		//				options when new reader is created to load the document. Once the reader is created
		//				changing theese properties while loading is in progress will not affect reader, to
		//				affect the reader in read events change the reader's object properties directly.
		// FormatOutput
		//				See help for this property on TXMLWriter object. This value here serve as a starting
		//				option when new writer is created to save the document. Once the writer is created
		//				changing that property while saving is in progress will not affect writer, to
		//				affect the writer in write events change the writer's object property directly.
		// ProgressStep
		//				See help for this property on TXMLReader object.
		property PreserveWhiteSpace: boolean read FPreserveWhiteSpace write FPreserveWhiteSpace default false;
		property MixedContent: boolean read FMixedContent write FMixedContent default false;
		property HTMLReader: boolean read FHTMLReader write FHTMLReader default false;
		property ProgressStep: integer read FProgressStep write FProgressStep default 5;
		property FormatOutput: boolean read FFormatOutput write FFormatOutput default true;

		// OnReadDocumentStart:
		//				Called just before reader starts to load xml document
		// OnReadDocumentEnd:
		//				Called just after reader finished loading of xml document
		// OnReadNode:
		//        Called whenever node is completed while reading. It's called for every single node
		//				in the document. Reader.Node represents the node that is read. If application frees
		//				the specified node then it will not be included in document after loading.
		//				Note: Freeing element node in this event will cause subsequent OnReadElementEnd
		//				with invalid Reader.Node object so care has to be taken if objects are freed here
		//				and OnReadElementEnd is processed.
		// OnReadElementStart:
		//        Called whenever element node is created while reading and attributes are parsed and
		//				stored as child nodes. Element is still empty since content is still not parsed.
		//				Reader.Node represents the element node that is started. Application can not free
		//				specified node because that would generate exception but it can change/free it's attributes
		//				since they are allready parsed in.
		// OnReadElementEnd:
		//        Called whenever element node is completed while reading. It will be called after OnReadNode
		//				for this same Node.	Reader.Node represents the element node that is completed. If application frees
		//				the specified node then it will not be included in document after loading.
		// OnReadProgress:
		//        Called whenever progress changes. Progress can be accessed through Reader.ProgressPercent
		//				and represents percent value from 0..100 of reading progress.
		//
		// Note:	if exception is raised in any of OnReadXXX events, then parsing is stopped, Document property will
		//				contain xml objects that are read up to that point and exception will be in raised state
		//				after LoadFromXXX method. No try/except blocks are implemented in XMLReader.
		//				If you want to allow user interaction while document is reading then call Application.ProceessMessages
		//				somewhere in OnReadXXX events.
		property OnReadDocumentStart: TXMLReaderEvent read FOnReadDocumentStart write FOnReadDocumentStart;
		property OnReadDocumentEnd: TXMLReaderEvent read FOnReadDocumentEnd write FOnReadDocumentEnd;
		property OnReadNode: TXMLReaderEvent read FOnReadNode write FOnReadNode;
		property OnReadElementStart: TXMLReaderEvent read FOnReadElementStart write FOnReadElementStart;
		property OnReadElementEnd: TXMLReaderEvent read FOnReadElementEnd write FOnReadElementEnd;
		property OnReadProgress: TXMLReaderEvent read FOnReadProgress write FOnReadProgress;

		// OnWriteDocumentStart:
		//				Called just before writer starts to write xml document
		// OnWriteDocumentEnd:
		//				Called just after writer finished writing of xml document
		// OnWriteElementStart:
		//        Called whenever element node is about to be written. Writer.Node represents the element node
		//				that is about to be written.
		// OnWriteElementEnd:
		//        Called whenever element node is completed while writing. Writer.Node represents the element
		//				node that is completed.
		// OnWriteProgress:
		//        Called whenever progress changes. Progress can be accessed through Writer.ProgressPercent
		//				and represents percent value from 0..100 of writing progress.
		property OnWriteDocumentStart: TXMLWriterEvent read FOnWriteDocumentStart write FOnWriteDocumentStart;
		property OnWriteDocumentEnd: TXMLWriterEvent read FOnWriteDocumentEnd write FOnWriteDocumentEnd;
		property OnWriteElementStart: TXMLWriterEvent read FOnWriteElementStart write FOnWriteElementStart;
		property OnWriteElementEnd: TXMLWriterEvent read FOnWriteElementEnd write FOnWriteElementEnd;
		property OnWriteProgress: TXMLWriterEvent read FOnWriteProgress write FOnWriteProgress;
	end;

const
	XMLPathSeparator: char = '/';

implementation

uses
	Math;

const
	CR: char = #13;
	LF: char = #10;
	TAB: char = #9;
	LetterSet: CharSet = ['A'..'Z', 'a'..'z'];
	WhitespaceSet: CharSet = [' ', #9, #10, #13];
	CombiningCharSet: CharSet = [];
	DigitSet: CharSet = ['0'..'9'];
	ExtenderSet: CharSet = [#$b7];
	FirstNameCharSet: CharSet = ['_', ':', 'A'..'Z', 'a'..'z'];
var
	NameCharSet: CharSet;
	NonQuotedAttrValueEndCharSet: CharSet;

const
	NodeNameJockerMany = '*';
	//NodeNameJockerOne = '?';
const
	HTMLNoEndTags: array [0..5] of string = (
		'META', 'IMG', 'LINK', 'HR', 'BR', 'INPUT'
	);

{ TXBuf }

procedure TXBuf.Init;
begin
	FSize := 0;
	FPtr := nil;
end;

procedure TXBuf.Clear;
begin
	if FPtr <> nil then
		begin
			//DebugInfo('TXBuf.Clear: FreeMem - before');
			FreeMem(FPtr);
			//DebugInfo('TXBuf.Clear: FreeMem - after');
			FPtr := nil;
		end;
	FSize := 0;
end;

procedure TXBuf.CopyFrom(SourcePtr: PChar; SourceSize: integer);
begin
	SetSize(SourceSize);
	if SourceSize > 0 then
		Move(SourcePtr^, FPtr^, SourceSize);
end;

procedure TXBuf.AppendFrom(SourcePtr: PChar; SourceSize: integer);
var
	FOldSize: integer;
begin
	if SourceSize > 0 then
		begin
			FOldSize := FSize;
			SetSize(FSize+SourceSize);
			Move(SourcePtr^, (FPtr+FOldSize)^, SourceSize);
		end;
end;

function TXBuf.GetAsString: string;
begin
	if FSize <= 0 then
		result := ''
	else
		begin
			SetLength(result, FSize);
			Move(FPtr^, result[1], FSize);
		end;
end;

procedure TXBuf.SetAsString(value: string);
begin
	Clear;
	FSize := Length(value);
	if FSize > 0 then
		begin
			GetMem(FPtr, FSize);
			Move(value[1], FPtr^, FSize);
		end;
end;

procedure TXBuf.SetSize(value: integer);
//var
//	FNewPtr: PChar;
begin
	if FSize <> value then
		begin
			//DebugInfo('TXBuf.SetSize: ReallocMem - before');
			//DebugInfoFmt('  Size: %d, Ptr: %s', [FSize, string(FPtr)]);

			(*
			if Value = 0 then
				Clear
			else
				begin
					GetMem(FNewPtr, value);
					if FSize > 0 then
						Move(FPtr^, FNewPtr^, Math.Min(FSize, value));
					if FPtr <> nil then
						FreeMem(FPtr);
					FPtr := FNewPtr;
					FSize := value;
				end;
			*)
			FSize := value;
			ReallocMem(FPtr, FSize);

			//DebugInfo('TXBuf.SetSize: ReallocMem - after');
		end;
end;

{:This function returns a string "i" characters long starting at the left
(beginning) of "AString".  Note that if "AString" is less that "i" characters
long this function will return the whole string. }
function Left(const AString: string; i: Integer): string;
begin
	Result := Copy(AString, 1, i);
end;

{:This function returns a string "i" characters long starting at the right
(end) of "AString".  Note that if "AString" is less that "i" characters
long this function will return the whole string. }
function Right(const AString: string; i: Integer): string;
begin
  if i > Length(AString) then
    Result := AString
  else
    Result := Copy(AString, Length(AString) - i + 1, i);
end;

(*
function TrimChars(const AString: string; TrimChars: CharSet): string;
var
	i, b, e: Integer;
	c: char;
	trim: boolean;
begin
	e := Length(AString);
	b := 1;
	trim := false;
	for i := 1 to e do
		begin
			c := AString[i];
			if c in TrimChars then
				begin
					b := i + 1;
					trim := true;
				end
			else
				Break;
		end;
	for i := e downto b do
		begin
			c := AString[i];
			if c in TrimChars then
				begin
					e := i - 1;
					trim := true;
				end
			else
				Break;
		end;
	if b > e then
		Result := ''
	else if trim then
		Result := Copy(AString, b, e - b + 1)
	else
		Result := AString;
end;
*)

function TrimWhiteSpace(const AString: string): string;
var
	i, b, e: Integer;
	trim: boolean;
begin
	e := Length(AString);
	b := 1;
	trim := false;
	for i := 1 to e do
		begin
			if AString[i] in WhiteSpaceSet then
				begin
					b := i + 1;
					trim := true;
				end
			else
				Break;
		end;
	for i := e downto b do
		begin
			if AString[i] in WhiteSpaceSet then
				begin
					e := i - 1;
					trim := true;
				end
			else
				Break;
		end;
	if b > e then
		Result := ''
	else if trim then
		Result := Copy(AString, b, e - b + 1)
	else
		Result := AString;
end;

function IsHTMLNoEndTag(TagName: string): boolean;
var
	i: byte;
begin
	result := false;
	for i := Low(HTMLNoEndTags) to High(HTMLNoEndTags) do
		begin
			if HTMLNoEndTags[i] = TagName then
				begin
					result := true;
					break;
				end;
		end;
end;

{ TXMLNode }

constructor TXMLNode.Create(const NodeType: TXMLNodeType);
begin
	inherited Create;
	FNodeType := NodeType;
end;

destructor TXMLNode.Destroy;
var
	n: TXMLNode;
begin
	//DebugInfo('TXMLNode.Destroy ' + Description);

	if FParentNode <> nil then
		FParentNode.RemoveChild(Self);

	while FLastChild <> nil do
		begin
			n := FLastChild;
			FLastChild := n.PreviousSibling;
			// Release the node
			n.FParentNode := nil;
			n.FPreviousSibling := nil;
			n.FNextSibling := nil;
			n.Free;
		end;

	inherited;
end;

procedure TXMLNode.SetParent(AParent: TXMLNode);
begin
	if FParentNode <> AParent then
		begin
			if FParentNode <> nil then
				FParentNode.RemoveChild(Self);
			if AParent <> nil then
				AParent.AppendChild(Self);
		end;
end;

function TXMLNode.GetOwnerDocument: TXMLNode;
begin
	if FNodeType = xntDocument then
		result := Self
	else if FParentNode = nil then
		result := nil
	else
		result := FParentNode.OwnerDocument;
end;

function TXMLNode.GetPath: string;
var
	n: TXMLNode;
begin
	result := '';
	n := ParentNode;
	while n <> nil do
		begin
			result := n.NodeName + XMLPathSeparator + result;
			n := n.ParentNode;
		end;
end;

function TXMLNode.Description: String;
begin
	Result := NodeName + ': ' + TXMLNodeTypeStrings[NodeType];
	if NodeValue <> '' then
		Result := Result + ' = ' + NodeValue;
end;

function TXMLNode.GetFirstChildNode(const NodeType: TXMLNodeType): TXMLNode;
begin
	result := FirstChild;
	while (result <> nil) and (result.NodeType <> NodeType) do
		result := result.NextSibling;
end;

function TXMLNode.GetNextSiblingNode(const NodeType: TXMLNodeType): TXMLNode;
begin
	result := NextSibling;
	while (result <> nil) and (result.NodeType <> NodeType) do
		result := result.NextSibling;
end;

function TXMLNode.GetTotalNodesCount: Integer;
var
	n: TXMLNode;
begin
	Result := 1;
	n := FFirstChild;
	while n <> nil do
		begin
			Result := Result + n.TotalNodesCount;
			n := n.FNextSibling;
		end;
end;

function TXMLNode.GetPreviousNode: TXMLNode;
begin
	result := PreviousSibling;
	if (result <> nil) and (result.LastChild <> nil) then
		result := result.LastChild;
	if result = nil then
		result := ParentNode;
end;

function TXMLNode.GetNextNode: TXMLNode;
begin
	result := FirstChild;
	if result = nil then
		result := NextSibling;
	if (result = nil) and (ParentNode <> nil) then
		result := ParentNode.NextSibling;
end;

function TXMLNode.CreateChild(const NodeType: TXMLNodeType): TXMLNode;
begin
	result := TXMLNode.Create(NodeType);
	AppendChild(result);
end;

procedure TXMLNode.AppendChild(Node: TXMLNode);
begin
	// if node belongs to other parent node remove the reference from the parent
	if Node.ParentNode <> nil then
		Node.ParentNode.RemoveChild(Node);
	Node.FParentNode := Self;

	if FLastChild <> nil then
		begin
			FLastChild.FNextSibling := Node;
			Node.FPreviousSibling := FLastChild;
		end
	else
		begin
			FFirstChild := Node;
		end;
	FLastChild := Node;
	Inc(FChildNodesCount);
end;

function TXMLNode.RemoveChild(Node: TXMLNode): boolean;
begin
	result := Node.FParentNode = self;

	if result then
		begin
			Node.FParentNode := nil;
			if FFirstChild = Node then
				FFirstChild := Node.FNextSibling;
			if FLastChild = Node then
				FLastChild := Node.FPreviousSibling;

			if Node.FNextSibling <> nil then
				begin
					Node.FNextSibling.FPreviousSibling := Node.FPreviousSibling;
					Node.FNextSibling := nil;
				end;
			if Node.FPreviousSibling <> nil then
				begin
					Node.FPreviousSibling.FNextSibling := Node.FNextSibling;
					Node.FPreviousSibling := nil;
				end;

			Dec(FChildNodesCount);
		end;
end;

function TXMLNode.HasChildNodes: boolean;
begin
  result := ChildNodesCount > 0;
end;

function TXMLNode.CloneNode(deep: boolean): TXMLNode;
var
	n: TXMLNode;
begin
	result := TXMLNode.Create(NodeType);
	try
		result.NodeName := NodeName;
		result.NodeValue := NodeValue;

		if deep then
      begin
				n := FirstChild;
				while n <> nil do
					begin
						result.AppendChild(n.CloneNode(true));
						n := n.NextSibling;
					end;
			end;

	except
		result.Free;
		raise;
	end;
end;

function TXMLNode.FindChild(const NodeName: String; const NodeType: TXMLNodeType = xntElement;
	const Index: Integer = 0; const CreateMissing: boolean = false): TXMLNode;
var
	ThisIndex: Integer;
	n: TXMLNode;
begin
	ThisIndex := -1;
	n := FFirstChild;
	while n <> nil do
		begin
			if (n.NodeType = NodeType) and (n.NodeName = NodeName) then
				begin
					Inc(ThisIndex);
					if ThisIndex = Index then
						begin
							Result := n;
							Exit;
						end;
				end;
			n := n.FNextSibling;
		end;

	Result := nil;

	if CreateMissing then
		begin
			while ThisIndex < Index do
				begin
					Result := CreateChild(NodeType);
					Result.NodeName := NodeName;
					Inc(ThisIndex);
				end;
		end;
end;

function TXMLNode.GetAttributes(Name: string): TXMLNode;
begin
	result := FindChild(Name, xntAttribute, 0, true);
end;

function TXMLNode.GetElements(Name: string): TXMLNode;
begin
	result := FindChild(Name, xntElement, 0, true);
end;

function TXMLNode.GetData: string;
var
	n: TXMLNode;
begin
	// To get Data value we have to add NodeValue with NodeValues of child text and CData nodes
	result := NodeValue;
	n := FirstChild;
	while n <> nil do
		begin
			if n.NodeType in [xntText, xntCData] then
				result := result + n.NodeValue;
			n := n.NextSibling;
		end;
end;

procedure TXMLNode.SetData(value: string);
var
	n, x: TXMLNode;
begin
	// To set Data value we have to delete child text and CData nodes and
	// put the value into NodeValue
	n := FirstChild;
	while n <> nil do
		begin
			if n.NodeType in [xntText, xntCData] then
				begin
					x := n.NextSibling;
					n.Free;
					n := x;
				end
			else
				n := n.NextSibling;
		end;
	NodeValue := value;
end;

function GetNodeText(Node: TXMLNode; Exact: boolean): string;
var
	n: TXMLNode;
	last_nt: TXMLNodeType;
	t: string;
begin
	// To get Text value we have to go along the object node tree structure
	// starting from this node and add all NodeValues for nodes of
	// xntElement, xntText or	xntCData type.
	result := Node.NodeValue;
	last_nt := xntDocument;
	n := Node.FirstChild;
	while n <> nil do
		begin
			if n.NodeType in [xntElement, xntText, xntCData] then
				begin
					t := GetNodeText(n, Exact);
					if not Exact and (n.NodeType = xntElement) and (last_nt = xntElement) and (result <> '') and (t <> '') then
						result := result + ' ' + t
					else
						result := result + t;
					last_nt := n.NodeType;
				end;
			n := n.NextSibling;
		end;
end;

function TXMLNode.GetText: string;
begin
	result := TrimWhiteSpace(GetNodeText(self, false));
end;

function TXMLNode.GetExactText: string;
begin
	result := TrimWhiteSpace(GetNodeText(self, true));
end;

function TXMLNode.GetXml: string;
var
	w: TXMLStringWriter;
begin
	result := '';
	w := TXMLStringWriter.Create(result);
	try
		w.StartDoc;
		w.WriteNode(Self);
		w.EndDoc;
	finally
		w.Free;
	end;
end;

(********************************************************************
 *
 *	Node selections
 *)

{$define TEST_PARSER0}

type
	TStepAxis = (
		AxRoot, AxParent, AxSelf, AxChild, AxAttribute, AxDescendant
	);
	TStep = record
		// Root is not realy an Axis, but we handle it that way
		Axis: TStepAxis;
		// only AxChild, AxAttribute, AxDescendant use name and index
		Name: string;
		Index: integer;
	end;
	TStepArray = array of TStep;
{$ifdef TEST_PARSER1}
const
	TStepAxisNames: array [TStepAxis] of string = (
		'AxRoot', 'AxParent', 'AxSelf', 'AxChild', 'AxAttribute', 'AxDescendant'
	);
{$endif TEST_PARSER1}

// ParseQuery compiles query string into step list for later processing
function ParseQuery(const queryString: String): TStepArray;
type
	TParseState = (stBeginOfStep, stNameStart, stNameChar, stIndex, stEndOfStep);
var
	nSteps: integer;
	NameStart: CharSet;
	NameChar: CharSet;
	state: TParseState;
	i: integer;
	c, next_c: char;

	procedure RaiseError(msg: string);
	begin
		raise Exception.CreateFmt(
			'selectSingleNode: Error at position %d of query string ''%s''. %s.',
			[i, queryString, msg]
		);
	end;

begin
	// Build charsets that names can consist of,
	// so that we know when parsing if illegal char is supplied in the position of the name
	NameStart := FirstNameCharSet + [NodeNameJockerMany];
	NameChar := NameCharSet + [NodeNameJockerMany];

	nSteps := 0;
	state := stBeginOfStep;
	i := 1;
	while i <= Length(queryString) do
		begin
			c := queryString[i];
			next_c := queryString[i+1];
			case state of

				stBeginOfStep:
				// New step is just started, processing the first char
					begin
						Inc(nSteps);
						SetLength(result, nSteps);
						if c = '.' then
							begin
								if next_c = '.' then
									begin
										// Parent node selection
										Inc(i); // eat next char
										result[nSteps-1].Axis := AxParent;
										state := stEndOfStep;
									end
								else
									begin
										// Self is selected
										result[nSteps-1].Axis := AxSelf;
										state := stEndOfStep;
									end;
							end
						else if c = '/' then
							begin
								// slash at the first position in a path denotes a root selection,
								// otherwise it is for descendants.
								if i = 1 then
									begin
										result[nSteps-1].Axis := AxRoot;
										// If there are more chars in the query, this belongs to the next step,
										// we don't require step separator here, as for normal step ending.
										if Length(queryString) > 1 then
											state := stBeginOfStep
										else
											state := stEndOfStep;
									end
								else
									begin
										result[nSteps-1].Axis := AxDescendant;
										state := stNameStart;
									end;
							end
						else if c = '@' then
							begin
								result[nSteps-1].Axis := AxAttribute;
								state := stNameStart;
							end
						else
							begin
								result[nSteps-1].Axis := AxChild;
								state := stNameStart;
								// Skip back one char, so that we again process is it in the name parsing state
								Dec(i);
							end;
					end;

				stNameStart:
				// Axis has been parsed, now we expect to start collecting node name until index or end of step
				// error is reported if improper char is found as the start of the node name
					begin
						if c in NameStart then
							begin
								result[nSteps-1].Name := c;
								result[nSteps-1].Index := 0;
								if i = Length(queryString) then
									state := stEndOfStep
								else
									state := stNameChar;
							end
						else
							RaiseError('Illegal start of node name');
					end;

				stNameChar:
				// we are collecting node name until index or end of step, or end of query
				// error is reported if improper char is found in the node name
					begin
						if c = '/' then
							begin
								state := stBeginOfStep;
							end
						else if c = '[' then
							begin
								// Index is present, we have to start it's parsing
								state := stIndex;
							end
						else if not (c in NameChar) then
							RaiseError('Illegal char ''' +c+ '''in node name')
						else
							begin
								result[nSteps-1].Name := result[nSteps-1].Name + c;
								if i = Length(queryString) then
									state := stEndOfStep;
							end;
					end;

				stIndex:
				// we are colecting index definition here, until ending ']'
					begin
						if c = ']' then
							begin
								state := stEndOfStep;
							end
						else if not (c in DigitSet) then
							RaiseError('Illegal char ''' +c+ '''in node index')
						else
							begin
								// Add digit to index
								result[nSteps-1].Index := result[nSteps-1].Index*10 + (Ord(c) - Ord('0'));
							end
					end;

				stEndOfStep:
				// Current step is just ended, waiting for step separator or end of query string
				// anything else is an error!
					begin
						if c = '/' then
							begin
								state := stBeginOfStep;
							end
						else
							RaiseError('Expecting char ''/'' or end of query');
					end;
			end;
			Inc(i);
		end;

	// ST_EndOfStep is only regular state that parsing can end with.
	if state <> stEndOfStep then
		RaiseError('Unexpected end of query');
end;

// ExecStep finds list of nodes that are result of applying location step ta a given node
function ExecStep(Node: TXMLNode; Step: TStep): TXMLNodeArray;
var
	NodeType: TXMLNodeType;
	NodeStart, NodeEnd: TXMLNode;
	FullSeek: boolean;
	i: integer;
begin
	SetLength(Result, 0);
	if Step.Axis in [AxRoot, AxParent, AxSelf] then
		begin
			// Root, Parent & Self axis don't support predicates
			// so there is only one resulting node from this
			case Step.Axis of
				AxRoot:		AddNodeToArray(Result, Node.OwnerDocument);
				AxParent:	AddNodeToArray(Result, Node.ParentNode);
				AxSelf:		AddNodeToArray(Result, Node);
			end;
		end
	else
		begin
			// Other axis can yield more than one node.
			// First we have to build list of nodes that satisfies axis and name,
			// and then filter those nodes to also satisfy index predicate if present.
			// What is left then is a list of nodes that meets this location step.
			// Note:
			//	code can be easily extended to support othe axis defined by XPath,
			//  we just have to properly define NodeStart, NodeEnd and FullSeek vars below.
			if Step.Axis = AxAttribute then
				NodeType := xntAttribute
			else
				NodeType := xntElement;
			NodeStart := Node.FirstChild;
			if Step.Axis <> AxDescendant then
				NodeEnd := Node.LastChild
			else
				begin
					NodeEnd := Node.NextSibling;
					if NodeEnd <> nil then
						NodeEnd := NodeEnd.PreviousNode; 
				end;
			FullSeek := Step.Axis = AxDescendant;
			while NodeStart <> nil do
				begin
					// Process node if matches type and name
					if (NodeStart.NodeType = NodeType) and NodeNameMatch(NodeStart.NodeName, Step.Name) then
						AddNodeToArray(Result, NodeStart);

					// If at the end break the loop, otherwise go for next node
					if NodeStart = NodeEnd then
						NodeStart := nil
					else if FullSeek then
						NodeStart := NodeStart.NextNode
					else
						NodeStart := NodeStart.NextSibling;
				end;
			// Now that the list of nodes is built, we have to filter out predicates if present
			if Step.Index <> 0 then
				begin
					for i := 0 to High(Result) do
						if i+1 <> Step.Index then
							Result[i] := nil;
					PackNodeArray(Result);
				end;
		end;
end;

function TXMLNode.selectNodes(const queryString: String): TXMLNodeArray;
var
	aSteps: TStepArray;
	iStep, iNode: integer;
	aNodesBefore, aNodesAfter, aStepNodes: TXMLNodeArray;
{$ifdef TEST_PARSER1}
	i: integer;
	s: string;
{$endif TEST_PARSER1}
begin
	// First is to parse the query string to break it down into individual steps
	aSteps := ParseQuery(queryString);
{$ifdef TEST_PARSER1}
	// for demonstration and testing only, we show the results of parsing
	s := '';
	for i := 0 to High(aSteps) do
		with aSteps[i] do
			s := s + Format('(Axis=%s, Name=%s, Index=%d)', [TStepAxisNames[Axis], Name, Index]);
	raise Exception.Create(s);
{$endif TEST_PARSER1}

	// For every location step we have to perform step execution,
	// as a starting point we go from the current node and for every step
	// transform every aNodesBefore into aNodesAfter. When last step is executed
	// result is returned.
	iStep := 0;
	AddNodeToArray(aNodesBefore, Self);
	while (Length(aNodesBefore) > 0) and (iStep <= High(aSteps)) do
		begin
			SetLength(aNodesAfter, 0);
			for iNode := 0 to High(aNodesBefore) do
				begin
					aStepNodes := ExecStep(aNodesBefore[iNode], aSteps[iStep]);
					AddNodesToArray(aNodesAfter, aStepNodes);
				end;

			Inc(iStep);
			aNodesBefore := aNodesAfter;
		end;
	result := aNodesBefore;
end;

function TXMLNode.selectSingleNode(const queryString: String): TXMLNode;
var
	aNodes: TXMLNodeArray;
begin
	aNodes := selectNodes(queryString);
	if Length(aNodes) = 0 then
		result := nil
	else
		result := aNodes[0];
end;

function NodeNameMatch(cName, cMatch: string): boolean;
var
	nNameLen, nMatchLen: byte;
	s_idx, m_idx, n: byte;
	m_ch: char;
begin
	nNameLen := Length(cName);
	nMatchLen := Length(cMatch);

	s_idx := 1;
	m_idx := 1;

	while (nNameLen > 0) and (nMatchLen > 0) do
		begin
			m_ch := cMatch[m_idx];
			Inc(m_idx);
			Dec(nMatchLen);

			(*if m_ch = NodeNameJockerOne then
				begin
					// ? zamjenjuje bilo koji znak
					Inc(s_idx);
					Dec(nNameLen);
				end
			else*) if m_ch = NodeNameJockerMany then
				begin
					// * replaces any sequence of chars

					// Support for NodeNameJockerOne right after the NodeNameJockerMany
					// is not done here! it should be included also, so that at least
					// so many chars must be skipped as the number of NodeNameJockerOne positions
					// after the NodeNameJockerMany.

					// n represents the length of match part after the * char that is of fixed chars.
					// (upto first jocker or the end of string).
					n := 0;
					while (n < nMatchLen) and
								//(cMatch[m_idx+n] <> NodeNameJockerOne) and
								(cMatch[m_idx+n] <> NodeNameJockerMany) do
						Inc(n);

					if (nMatchLen = 0) or (n = 0) then
						begin
							// Jocker at the end replaces all of the rest string characters
							nNameLen := 0
						end
					else
						while (nNameLen >= n) and (Copy(cName, s_idx, n) <> Copy(cMatch, m_idx, n)) do
							begin
								Inc(s_idx);
								Dec(nNameLen);
							end;
				end
			else if cName[s_idx] <> m_ch then
				begin
					// Uzorci nisu jednaki
					Result := False;
					Exit;
				end
			else
				begin
					// Uzorci su jednaki pa preskace taj znak u stringu
					Inc(s_idx);
					Dec(nNameLen);
				end;
		end;

	// Skip all the unmatched jockers from the end.
	while (nMatchLen > 0) and (cMatch[m_idx] = NodeNameJockerMany) do
		begin
			Inc(m_idx);
			Dec(nMatchLen);
		end;

	// If both strings matched then ok
	Result := (nMatchLen = 0) and (nNameLen = 0);
end;

procedure AddNodeToArray(var aNodes: TXMLNodeArray; Node: TXMLNode);
var
	n: integer;
begin
	n := Length(aNodes);
	SetLength(aNodes, n+1);
	aNodes[n] := Node;
end;

procedure AddNodesToArray(var aNodes: TXMLNodeArray; AddNodes: TXMLNodeArray);
var
	n1, n2, i: integer;
begin
	n1 := Length(aNodes);
	n2 := Length(AddNodes);
	SetLength(aNodes, n1+n2);
	for i := 0 to n2-1 do
		aNodes[n1+i] := AddNodes[i];
end;

procedure PackNodeArray(var aNodes: TXMLNodeArray);
var
	i, n: integer;
	NewNodes: TXMLNodeArray;
begin
	n := 0;
	for i := 0 to High(aNodes) do
		if aNodes[i] <> nil then
			Inc(n);
	if n < Length(aNodes) then
		begin
			// Array needs repacking
			SetLength(NewNodes, n);
			n := 0;
			for i := 0 to High(aNodes) do
				if aNodes[i] <> nil then
					begin
						NewNodes[n] := aNodes[i];
						Inc(n);
					end;
			aNodes := NewNodes;
		end;
end;

function IsNodeInArray(var aNodes: TXMLNodeArray; Node: TXMLNode): boolean;
var
	i: integer;
begin
	result := false;
	for i := 0 to High(aNodes) do
		if aNodes[i] = Node then
			begin
				result := true;
				break;
			end;
end;

{ TXMLReader }

(*
function TXMLReader.ReadBufOpen: integer;
begin
	result := FReadBufPos;
end;
*)

procedure TXMLReader.ReadBufAdd(c: char);
begin
	if FReadBufPos >= FReadBuf.Size then
		FReadBuf.Size := FReadBuf.Size + 256;

	FReadBuf.Ptr[FReadBufPos] := c;
	Inc(FReadBufPos);
end;

function TXMLReader.ReadBufClose(p: integer): TXBuf;
begin
	Result.Init;
	if p < FReadBufPos then
		begin
			Result.FSize := FReadBufPos - p;
			Result.FPtr := FReadBuf.Ptr + p;
			FReadBufPos := p;
		end;
end;

procedure TXMLReader.UnReadChar(c: char);
begin
	if FUnReadBufPos >= FUnReadBuf.Size then
		FUnReadBuf.Size := FUnReadBuf.Size + 256;

	FUnReadBuf.Ptr[FUnReadBufPos] := CurChar;
	Inc(FUnReadBufPos);

	CurChar := c;
end;

procedure TXMLReader.UnReadString(p: PChar; count: integer);
begin
	while count > 0 do
		begin
			Dec(count);
			UnReadChar(p[count]);
		end;
end;

procedure TXMLReader.ReadChar;
begin
	if FUnReadBufPos > 0 then
		begin
			// There are characters in the unread buffer, so we get it from there
			Dec(FUnReadBufPos);
			CurChar := FUnReadBuf.Ptr[FUnReadBufPos];
		end
	else
		begin
			CurChar := GetChar;
			Inc(FColNumber);
			// Line & Col numbering maintenance
			case FLineState of
				0:	begin
							if CurChar = CR then
								FLineState := 1
							else if CurChar = LF then
								begin
									Inc(FLineNumber);
									FColNumber := 0;
								end;
						end;
				1:	begin
							if CurChar = LF then
								begin
									Inc(FLineNumber);
									FColNumber := 0;
								end;
							FLineState := 0;
						end;
			end;

			// Progress indicator support only if we read from the XML source
			// because progress is based on that not the content of unread buffer
			// which may be filled out by the application in certain circumstances
			// when we realize the parsing that way.
			Dec(FProgressPos);
			if FProgressPos = 0 then
				begin
					Inc(FProgressPercent, FProgressStep);
					FProgressPos := FProgressCount;
					DoProgress;
				end;
		end;
end;

procedure TXMLReader.ReadCharAndCheckForEnd;
begin
	if EndOfData then
		ReportError('Unexpected end of data');
	ReadChar;
end;

procedure TXMLReader.DoNode(Node: TXMLNode);
begin
	if Assigned(FOnNode) then
		begin
			FNode := Node;
			FOnNode(Self);
		end;
end;

procedure TXMLReader.DoElementStart(Node: TXMLNode);
begin
	if Assigned(FOnElementStart) then
		begin
			FNode := Node;
			FOnElementStart(Self);
		end;
end;

procedure TXMLReader.DoElementEnd(Node: TXMLNode);
begin
	if Assigned(FOnElementEnd) then
		begin
			FNode := Node;
			FOnElementEnd(Self);
		end;
end;

procedure TXMLReader.DoProgress;
begin
	if Assigned(FOnProgress) then	FOnProgress(Self);
end;

function TXMLReader.DoesStreamMatchString(Str: String; CaseSensitive: boolean = true): Boolean;
var
	i: Integer;
	c: char;
begin
	if Str = '' then
		Result := False
	else
		begin
			i := 1;
			repeat
				c := CurChar;
				if CaseSensitive then
					Result := Str[i] = CurChar
				else
					Result := UpperCase(Str[i]) = UpperCase(CurChar);
				Inc(i);
				ReadChar;
			until (not Result) or EndOfData or (i > Length(Str));

			if not Result then
				begin
					UnReadChar(c);
					UnReadString(@Str[1], i-2);
				end;
		end;
end;

procedure TXMLReader.ParseAttribute(Node: TXMLNode);

	function ParseAttValue(var AttValue: String): Boolean;
	var
		QuoteChar: Char;
	begin
		// We allow attribute values with or without quotes
		// to make easy of parsing HTML files that usualy goes without quotes.
		Result := False;

		if CurChar in ['"', ''''] then
			begin
				QuoteChar := CurChar;
				ReadCharAndCheckForEnd;
			end
		else if FHTMLReader then
			begin
				QuoteChar := #0;
			end
		else
			ReportError('A string literal was expected, but no opening quote character was found.');

		while not Result do
			begin
				if ((QuoteChar <> #0) and (CurChar = QuoteChar)) or
					 ((QuoteChar = #0) and (CurChar in NonQuotedAttrValueEndCharSet)) then
					Result := True
				else
					begin
						if CurChar = '<' then
							ReportError('The character ''' + CurChar + ''' cannot be used in an attribute value')
						else if CurChar = '&' then
							CurChar := ParseReference;
						AttValue := AttValue + CurChar;
						ReadCharAndCheckForEnd;
					end;
			end;

		if QuoteChar <> #0 then
			ReadCharAndCheckForEnd
		else
			begin
				// We have to check if value ended with '/' and current char is '>'
				// If this is so we have to unread '/' from argument value.
				// Unquoted values should read args like type=text/javascript
				// so we can not just break on '/' char, but /> should be treated differently.
				if (CurChar = '>') and (AttValue[Length(AttValue)] = '/') then
					begin
						UnReadChar(CurChar);
						SetLength(AttValue, Length(AttValue)-1);
					end;
			end;
	end;

var
	Name: TXBuf;
	HasValue: boolean;
	AttValue: String;
	ThisNode: TXMLNode;
begin
	ParseName(Name);
	ThisNode := Node.CreateChild(xntAttribute);
	ThisNode.NodeName := Name.AsString;
	HasValue := ParseEq;
	if not HTMLReader and not HasValue then
		ReportError('Missing equals sign between attribute and attribute value.');
	if HasValue then
		begin
			if not ParseAttValue(AttValue) then
				ReportError('Attribute ''' + ThisNode.NodeName + ''' is missing value. A string literal was expected, but no opening quote character was found');
			ThisNode.NodeValue := AttValue;
		end;
	DoNode(ThisNode);
end;

(*
function TXMLReader.ParseAttValue(var AttValue: String): Boolean;
var
	QuoteChar: Char;
begin
	Result := False;
	if not (CurChar in ['"', '''']) then Exit;
	QuoteChar := CurChar;
	ReadCharAndCheckForEnd;

	while CurChar <> QuoteChar do
		begin
			if CurChar = '<' then
				ReportError('The character ''' + CurChar + ''' cannot be used in an attribute value')
			else if CurChar = '&' then
				CurChar := ParseReference;
			AttValue := AttValue + CurChar;
			ReadCharAndCheckForEnd;
		end;

	ReadCharAndCheckForEnd;
	Result := True;
end;
*)

procedure TXMLReader.ParseComment;
var
	Comment: String;
	Terminated: Boolean;
begin
	repeat
		Comment := Comment + CurChar;
		Terminated := Right(Comment, 3) = '-->';
		if EndOfData then Break;
		ReadChar;
	until Terminated;
	if Right(Comment, 4) = '--->' then
		ReportError('Incorrect syntax was used in a comment. Comments cannot end with ''--->''');
end;

procedure TXMLReader.ParseContent(Node: TXMLNode; EndTagName: string);

var
	p: integer;
	Data: TXBuf;

const
	CDataEndMax = 3;
var
	CDataEndPos: byte;

	procedure ParseCDSect;
	var
		FReadBufFreeze: integer;
		CData: TXBuf;
	begin
		FReadBufFreeze := FReadBufPos;
		CDataEndPos := 0;
		repeat
			ReadBufAdd(CurChar);
			if CurChar = ']' then
				begin
					if CDataEndPos < (CDataEndMax-1) then
						Inc(CDataEndPos);
				end
			else if (CurChar = '>') and (CDataEndPos = CDataEndMax-1) then
				begin
					Inc(CDataEndPos);
				end
			else
				CDataEndPos := 0;
			ReadCharAndCheckForEnd;
		until CDataEndPos = CDataEndMax;

		// Remove CDATA ending sequence
		Dec(FReadBufPos, CDataEndMax);

		if FMixedContent then
			begin
				CData := ReadBufClose(FReadBufFreeze);
				with Node.CreateChild(xntCData) do
					begin
						NodeValue := CData.AsString;
					end;
			end;
	end;

	procedure ParseCharData;
	var
		bWhiteSpace: boolean;
		FReadBufFreeze: integer;
		CharData: TXBuf;
	begin
		FReadBufFreeze := FReadBufPos;
		bWhiteSpace := not FPreserveWhiteSpace;
		CDataEndPos := 0;
		repeat
			if CurChar = '&' then
				begin
					ReadBufAdd(ParseReference);
					bWhiteSpace := false;
					CDataEndPos := 0;
				end
			else
				begin
					ReadBufAdd(CurChar);
					if bWhiteSpace and not (CurChar in WhiteSpaceSet) then
						bWhiteSpace := false;
					if CurChar = ']' then
						begin
							if CDataEndPos < (CDataEndMax-1) then
								Inc(CDataEndPos);
						end
					else if (CurChar = '>') and (CDataEndPos = CDataEndMax-1) then
						ReportError('The literal string '']]>'' is not allowed in element content')
					else
						CDataEndPos := 0;
					ReadChar;
				end;
		until CurChar in ['<', #0];

		// If char data contains only white space and if we don't preserve it
		// then we drop it, otherwise we just append it to the content data buffer
		if bWhiteSpace then
			begin
				FReadBufPos := FReadBufFreeze;
			end
		else if FMixedContent then
			begin
				CharData := ReadBufClose(FReadBufFreeze);
				with Node.CreateChild(xntText) do
					begin
						NodeValue := CharData.AsString;
					end;
			end;
	end;

	function ParseETag: boolean;
	var
		Name: TXBuf;
	begin
		ParseName(Name);
		// HTML tags are case insensitive, while XML tags are not, so we have to take care of that
		if not HTMLReader then
			Result := Name.AsString = EndTagName
		else
			Result := UpperCase(Name.AsString) = UpperCase(EndTagName);
		if Result then
			begin
				ParseWhitespace;
				if CurChar <> '>' then
					ReportError('The character ''>'' was expected.');
				ReadChar;
			end
		else
			begin
				if not HTMLReader then
					ReportError('End tag ''' + Name.AsString + ''' does not match the start tag ''' + EndTagName + '''')
				else
					begin
						// For HTML parsing we allow omitting close tags for nodes,
						// also any close tags that are without open tags are ignored.
						if Pos(UpperCase(Name.AsString)+XMLPathSeparator, UpperCase(Node.Path)) = 0 then
							begin
								// Ignore close tag, and position reader after it
								ParseWhitespace;
								if CurChar = '>' then
								ReadChar;
							end
						else
							begin
								// Unread close tag so that parent content parsing can parse it and close it's node regularly
								UnReadString(Name.Ptr, Name.Size);
								UnReadChar('/');
								UnReadChar('<');
								Result := true;
							end;
					end;
			end;
	end;

var
	FoundContent: Boolean;
begin
	p := FReadBufPos;

	FoundContent := False;
	while not FoundContent do
		begin
			if CurChar = '<' then
				begin
					ReadCharAndCheckForEnd;
					case CurChar of
						'/':
							begin
								ReadCharAndCheckForEnd;
								FoundContent := ParseETag;
							end;
						'?':
							begin
								ReadCharAndCheckForEnd;
								ParseProcessingInstruction(Node);
							end;
						'!':
							begin
								ReadCharAndCheckForEnd;
								if DoesStreamMatchString('--') then
									ParseComment
								else if DoesStreamMatchString('[CDATA[') then
									ParseCDSect
								else
									ReportError('Unrecognized character sequence after <!');
							end;
						else
							ParseElement(Node);
					end;
				end
			else if CurChar = #0 then
				ReportError('Unexpected end of character stream')
			else
				ParseCharData;
		end;
	Data := ReadBufClose(p);
	if Data.Size > 0 then
		begin
			Node.NodeValue := Data.AsString;
		end;
end;

procedure TXMLReader.ParseContentAsIs(var Content: TXBuf; EndSequence: string; CaseSensitive: boolean = true);
var
	p: integer;
begin
	p := FReadBufPos;
	while true do
		begin
			if (CurChar = EndSequence[1]) and DoesStreamMatchString(EndSequence, CaseSensitive) then
				break
			else
				begin
					ReadBufAdd(CurChar);
					ReadCharAndCheckForEnd;
				end;
		end;
	Content := ReadBufClose(p);
end;

constructor TXMLReader.Create;
begin
	FPreserveWhiteSpace := false;
	FMixedContent := false;
	FHTMLReader := false;
	FProgressStep := 5;
end;

destructor TXMLReader.Destroy;
begin
	FReadBuf.Clear;
	FUnReadBuf.Clear;
	inherited;
end;

procedure TXMLReader.ParseDocument(Document: TXMLNode);

	procedure ParseXMLDecl;
	var
		ThisNode: TXMLNode;
		IsDone: boolean;
	begin
		if DoesStreamMatchString('<?xml ') then
			begin
				ThisNode := Document.CreateChild(xntXMLDeclaration);
				ThisNode.NodeName := 'xml';
				IsDone := false;
				while not IsDone do
					begin
						ParseWhitespace;
						if CurChar = '?' then
							IsDone := true
						else if CurChar = '>' then
							ReportError('Invalid syntax for an xml declaration.')
						else
							ParseAttribute(ThisNode);
					end;
				ReadCharAndCheckForEnd;
				if CurChar <> '>' then
					ReportError('Invalid syntax for an xml declaration.');
				ReadChar;
				DoNode(ThisNode);
			end;
	end;

	procedure ParseDocTypeDecl;
	var
		ThisNode: TXMLNode;
		Name, Data: TXBuf;
		p: integer;
		OpenCount: integer;
	begin
		if DoesStreamMatchString('<!DOCTYPE') then
			begin
				ParseWhitespace;
				ParseName(Name);

				ThisNode := Document.CreateChild(xntDocumentType);
				ThisNode.NodeName := Name.AsString;
				ParseWhitespace;

				OpenCount := 1;
				// Document type data is stored as-is, no parsing, just collect the text until the ending >
				p := FReadBufPos;
				repeat
					if CurChar = '>' then
						Dec(OpenCount)
					else if CurChar = '<' then
						Inc(OpenCount);
					if OpenCount > 0 then
						begin
							ReadBufAdd(CurChar);
						end;
					ReadCharAndCheckForEnd;
				until OpenCount = 0;
				Data := ReadBufClose(p);

				ThisNode.NodeValue := Data.AsString;
				DoNode(ThisNode);
			end;
	end;

	function ParseMisc: Boolean;
	begin
		Result := true;
		if ParseWhitespace then Exit;
		if DoesStreamMatchString('<!--') then
			begin
				ParseComment;
				Exit;
			end;
		if DoesStreamMatchString('<?') then
			begin
				ParseProcessingInstruction(Document);
				Exit;
			end;
		Result := false;
	end;

	function ParseProlog: Boolean;
	begin
		ParseWhitespace;
		ParseXMLDecl;
		while ParseMisc do
			;
		ParseDocTypeDecl;
		while ParseMisc do
			;
		Result := True;
	end;

begin
	//DebugInfo('ParseDocument Begin');
	FProgressPercent := 0;
	FProgressCount := (Size div 100) * FProgressStep;
	FProgressPos := FProgressCount;
	FLineNumber := 1;
	FColNumber := 0;
	FLineState := 0;
	DoProgress;

	ReadCharAndCheckForEnd;
	if not ParseProlog then
		ReportError('Missing prolog');

	if CurChar <> '<' then
		ReportError('Missing root element');
	ReadCharAndCheckForEnd;
	ParseElement(Document);

	while ParseMisc do
		;
	//DebugInfo('ParseDocument End');
	//DebugInfoFmt('  FSetNameCount = %d, FSetDataCount = %d', [FSetNameCount, FSetDataCount]);

	FProgressPercent := 100;
	DoProgress;
end;

procedure TXMLReader.ParseElement(Node: TXMLNode);
var
	Name: TXBuf;
	ThisNode: TXMLNode;
	IsDone: boolean;
begin
	ParseName(Name);

	ThisNode := Node.CreateChild(xntElement);
	ThisNode.NodeName := Name.AsString;
	IsDone := false;
	while not IsDone do
		begin
			ParseWhitespace;
			if CurChar in ['/', '>'] then
				IsDone := true
			else if HTMLReader and (CurChar = '<') then
				begin
					// For HTML processing we allow for tags to be left unclosed,
					// and close them automaticaly
					UnReadChar('<');
					CurChar := '>';
					IsDone := true;
				end
			else
				ParseAttribute(ThisNode);
		end;

	if CurChar = '/' then begin
		// EmptyElemTag
		ReadCharAndCheckForEnd;
		if CurChar <> '>' then
			ReportError('The character ''>'' was expected.');
		ReadChar;
		DoElementStart(ThisNode);
	end else if CurChar = '>' then begin
		// STag
		ReadCharAndCheckForEnd;
		DoElementStart(ThisNode);
		if not FHTMLReader then
			ParseContent(ThisNode, ThisNode.NodeName)
		else
			begin
				if UpperCase(ThisNode.NodeName) = 'SCRIPT' then
					begin
						ParseContentAsIs(Name, '</SCRIPT>', false);
						ThisNode.NodeValue := Name.AsString;
					end
				else if not IsHTMLNoEndTag(UpperCase(ThisNode.NodeName)) then
					ParseContent(ThisNode, ThisNode.NodeName);
			end;
	end;

	DoNode(ThisNode);
	DoElementEnd(ThisNode);
end;

function TXMLReader.ParseEq: Boolean;
begin
	ParseWhitespace;
	Result := CurChar = '=';
	if Result then
		begin
			ReadCharAndCheckForEnd;
			ParseWhitespace;
		end;
end;

procedure TXMLReader.ParseName(var Name: TXBuf);
var
	p: integer;
begin
	if CurChar in FirstNameCharSet then
		begin
			p := FReadBufPos;
			repeat
				ReadBufAdd(CurChar);
				ReadCharAndCheckForEnd;
			until not (CurChar in NameCharSet);
			Name := ReadBufClose(p);
		end
	else if CurChar in WhiteSpaceSet then
		ReportError('Whitespace is not allowed at this location.')
	else
		ReportError('A name was started with an invalid character.');
end;

procedure TXMLReader.ParseProcessingInstruction(Node: TXMLNode);
var
	Name: TXBuf;
	Chars: string;
	ThisNode: TXMLNode;
	Terminated: Boolean;
begin
	ParseName(Name);
	if Name.AsString = 'xml' then
		ReportError('Processing instruction cannot have the reserved name "xml"');
	ThisNode := Node.CreateChild(xntProcessingInstruction);
	ThisNode.NodeName := Name.AsString;
	ParseWhitespace;
	repeat
		Chars := Chars + CurChar;
		Terminated := Right(Chars, 2) = '?>';
		if EndOfData then break;
		ReadChar;
	until Terminated;
	if not Terminated then
		ReportError('Unterminated processing instruction');
	ThisNode.NodeValue := Left(Chars, Length(Chars) - 2);
	
	DoNode(ThisNode);
end;

function TXMLReader.ParseReference: Char;
var
	RefStr: String;
begin
	Result := #0; // silence compiler warning
	if CurChar <> '&' then
		ReportError('All references must begin with "&"');
	ReadCharAndCheckForEnd;
	while CurChar <> ';' do begin
		RefStr := RefStr + CurChar;
		ReadCharAndCheckForEnd;
	end;
	if Left(RefStr, 2) = '#x' then
		try
			Result := Chr(StrToInt('$' + Right(RefStr, Length(RefStr) - 2)));
		except
			ReportError('Invalid integer in reference &' + RefStr + ';');
		end
	else if Left(RefStr, 1) = '#' then
		try
			Result := Chr(StrToInt(Right(RefStr, Length(RefStr) - 1)));
		except
			ReportError('Invalid integer in reference &' + RefStr + ';');
		end
	else if RefStr = 'amp' then
		Result := '&'
	else if RefStr = 'lt' then
		Result := '<'
	else if RefStr = 'gt' then
		Result := '>'
	else if RefStr = 'quot' then
		Result := '"'
	else if RefStr = 'apos' then
		Result := ''''
	else if HTMLReader and (RefStr = 'nbsp') then
		Result := ' '
	else
		ReportError('Reference to undefined entity '''+RefStr+'''.');

	ReadCharAndCheckForEnd;
end;

function TXMLReader.ParseWhitespace: Boolean;
begin
	Result := False;
	while (CurChar in WhitespaceSet) do
		begin
			Result := True;
			ReadChar;
			if EndOfData then
				Exit;
		end;
end;

procedure TXMLReader.ReportError(Err: String);
begin
	raise Exception.CreateFmt('%s' + #13#10 + 'Line %d, Position %d', [Err, LineNumber, ColNumber]);
end;

{ TXMLStringReader }

constructor TXMLStringReader.Create(AData: string);
begin
	inherited Create;
	FData := AData;
	FPosition := 0;
end;

function TXMLStringReader.GetSize: integer;
begin
	result := length(FData);
end;

function TXMLStringReader.GetPosition: integer;
begin
	result := FPosition;
end;

function TXMLStringReader.EndOfData: Boolean;
begin
	Result := FPosition >= Length(FData);
end;

function TXMLStringReader.GetChar: char;
begin
	if FPosition >= Length(FData) then
		result := #0
	else
		begin
			Inc(FPosition);
			result := FData[FPosition];
		end
end;

{ TXMLStreamReader }

constructor TXMLStreamReader.Create(AStream: TStream);
begin
	inherited Create;
	FCacheLen := 0;
	FBehind := 0;
	FStream := AStream;
	FSize := AStream.Size;
	FPosition := AStream.Position;
end;

function TXMLStreamReader.GetSize: integer;
begin
	result := FSize;
end;

function TXMLStreamReader.GetPosition: integer;
begin
	result := FPosition - FBehind;
end;

function TXMLStreamReader.EndOfData: Boolean;
begin
	Result := (FPosition - FBehind) >= FSize;
end;

function TXMLStreamReader.GetChar: char;
begin
	if FBehind > 0 then
		begin
			result := FCacheBuf[FCacheLen-FBehind];
			Dec(FBehind);
		end
	else
		begin
			if FPosition >= FSize then
				result := #0
			else
				begin
					// Load cache as much as posible
					FCacheLen := FStream.Read(FCacheBuf[0], XMLStreamReaderCacheSize);
					FBehind := FCacheLen;
					FPosition := FStream.Position;

					result := FCacheBuf[FCacheLen-FBehind];
					Dec(FBehind);
				end;
		end;
end;

{ TXMLFileReader }

constructor TXMLFileReader.Create(AFileName: string);
begin
	inherited Create(TFileStream.Create(AFileName, fmOpenRead));
	FFileName := AFileName;
end;

destructor TXMLFileReader.Destroy;
begin
	if Stream <> nil then
		Stream.Free;
	inherited;
end;

{ TXMLWriter }

constructor TXMLWriter.Create;
begin
	inherited;
	FFormatOutput := true;
	//FProgressStep := 5;
end;

destructor TXMLWriter.Destroy;
begin
	FWriteBuf.Clear;
	inherited;
end;

procedure TXMLWriter.StartDoc;
begin
	FWriteBuf.Size := XMLWriterCacheSize;
	FWriteBufPos := 0;
	FIndentation := 0;
	FFirstLine := true;

	FPosition := 0;
	FTotal := 0;
	FCount := 0;
	FProgressPercent := 0;
end;

procedure TXMLWriter.EndDoc;
begin
	if FWriteBufPos > 0 then
		WriteBuffer(FWriteBuf.Ptr, FWriteBufPos);
	FWriteBuf.Clear;
end;

procedure TXMLWriter.WriteBuffer(Buffer: PChar; Count: integer);
begin
	// Progress count support
	Inc(FPosition, Count);
	if FTotal <> 0 then
		FProgressPercent := (100 * FCount) div FTotal;
	DoProgress;
end;

procedure TXMLWriter.WriteChar(c: char);
begin
	if FWriteBufPos >= XMLWriterCacheSize then
		begin
			WriteBuffer(FWriteBuf.Ptr, FWriteBufPos);
			FWriteBufPos := 0;
		end;
	FWriteBuf.Ptr[FWriteBufPos] := c;
	Inc(FWriteBufPos);
end;

procedure TXMLWriter.DoElementStart(Node: TXMLNode);
begin
	if Assigned(FOnElementStart) then
		begin
			FNode := Node;
			FOnElementStart(Self);
		end;
end;

procedure TXMLWriter.DoElementEnd(Node: TXMLNode);
begin
	if Assigned(FOnElementEnd) then
		begin
			FNode := Node;
			FOnElementEnd(Self);
		end;
end;

procedure TXMLWriter.DoProgress;
begin
	if Assigned(FOnProgress) then	FOnProgress(Self);
end;

procedure TXMLWriter.WriteStr(const Value: String);
var
	done: boolean;
	Size: integer;
begin
	done := false;
	Size := Length(Value);
	while not done do
		if FWriteBufPos + Size <= XMLWriterCacheSize then
			begin
				Move(Value[1], FWriteBuf.Ptr[FWriteBufPos], Size);
				Inc(FWriteBufPos, Size);
				done := true;
			end
		else if FWriteBufPos > 0 then
			begin
				WriteBuffer(FWriteBuf.Ptr, FWriteBufPos);
				FWriteBufPos := 0;
			end
		else
			begin
				WriteBuffer(@Value[1], Size);
				done := true;
			end;
end;

procedure TXMLWriter.Write(const Value: String);
var
	Size: integer;
	i: Integer;
	c: char;
begin
	Size := Length(Value);
	i := 1;
	while i <= Size do
		begin
			c := Value[i];
			case c of
				'<': 	WriteStr('&lt;');
				'>': 	WriteStr('&gt;');
				'''':	WriteStr('&apos;');
				'"': 	WriteStr('&quot;');
				'&':	WriteStr('&amp;');
			else
				(*if Ord(c) > 127 then
					WriteStr('&#' + IntToStr(Ord(Value[i])) + ';')
				else*)
					WriteChar(c);
			end;
			Inc(i);
		end;
end;

procedure TXMLWriter.WriteLineEnd;
var
	i: integer;
begin
	if FFormatOutput then
		begin
			if FFirstLine then
				FFirstLine := false
			else
				begin
					WriteChar(#13);
					WriteChar(#10);
					for i := 1 to FIndentation do
						begin
							WriteChar(' ');
							WriteChar(' ');
						end;
				end
		end;
end;

procedure TXMLWriter.WriteStandaloneDocumentDeclaration;
begin
	WriteStr('<?xml version="1.0" standalone="yes"?>');
end;

procedure TXMLWriter.WriteTag(const Name: String; TagType: TXMLTagType;
	Open: Boolean);
begin
	if TagType in [xttStarting, xttStartingAndEnding] then
		begin
			WriteLineEnd;
			WriteChar('<');
		end
	else if TagType in [xttEnding, xttEndingSameLine] then
		begin
			Dec(FIndentation);
			if TagType = xttEnding then
				WriteLineEnd;
			WriteStr('</');
		end;
	WriteStr(Name);
	if not Open then
		WriteTagClose(TagType);
end;

procedure TXMLWriter.WriteTagClose(TagType: TXMLTagType);
begin
	if TagType = xttStartingAndEnding then
		WriteStr('/>')
	else
		begin
			WriteStr('>');
			if TagType = xttStarting then
				Inc(FIndentation);
		end;
end;

procedure TXMLWriter.WriteCDATA(Data: string);
const
	CDataEndMax = 3;
var
	CDataEndPos: byte;
	i, size: integer;
	c: char;
begin
	size := Length(Data);
	if size > 0 then
		begin
			CDataEndPos := 1;
			WriteStr('<![CDATA[');
			for i := 1 to Size do
				begin
					c := Data[i];
					if c = ']' then
						begin
							if CDataEndPos < (CDataEndMax-1) then
								Inc(CDataEndPos);
							WriteChar(c);
						end
					else if (c = '>') and (CDataEndPos = CDataEndMax-1) then
						begin
							CDataEndPos := 0;
							WriteStr(']]>');
							Write(c);
							WriteStr('<![CDATA[');
						end
					else
						begin
							CDataEndPos := 0;
							WriteChar(c);
						end;
				end;
			WriteStr(']]>');
		end;
end;

procedure TXMLWriter.WriteData(const Name, Data: String);
begin
	//if Value = '' then Exit;
	WriteTag(Name);
	Write(Data);
	WriteTag(Name, xttEndingSameLine);
end;

procedure TXMLWriter.WriteTagParam(const ParamName, Value: String);
begin
	WriteChar(' ');
	WriteStr(ParamName);
	WriteChar('=');
	WriteChar('"');
	Write(Value);
	WriteChar('"');
end;

procedure TXMLWriter.WriteNode(Node: TXMLNode);
var
	is_empty: boolean;
	has_elem: boolean;
	n: TXMLNode;
begin
	// If WriteNode called immediately after StartDoc
	// then we have to set total nodes count so that
	// progress can be maintained
	if FTotal = 0 then
		FTotal := Node.TotalNodesCount;

	if Node.NodeType = xntXMLDeclaration then
		begin
			WriteLineEnd;
			WriteStr('<?xml');
			n := Node.FirstChild;
			while n <> nil do
				begin
					if n.NodeType = xntAttribute then
						WriteTagParam(n.NodeName, n.NodeValue);
					n := n.NextSibling;
					Inc(FCount);
				end;
			WriteStr('?>');
		end
	else if Node.NodeType = xntProcessingInstruction then
		begin
			WriteLineEnd;
			WriteStr('<?' + Node.NodeName + ' ' + Node.NodeValue + '?>');
		end
	else if Node.NodeType in [xntDocument, xntElement] then
		begin
			if Node.NodeType = xntElement then
				DoElementStart(Node);

			is_empty := Node.NodeValue = '';
			has_elem := Node.NodeType = xntDocument;

			if Node.NodeType = xntElement then
				begin
					WriteTag(Node.NodeName, xttStarting, true);

					n := Node.FirstChild;
					while n <> nil do
						begin
							if (n.NodeType = xntAttribute) then
								begin
									WriteTagParam(n.NodeName, n.NodeValue);
									Inc(FCount);
								end
							else if n.NodeType in [xntText, xntCData] then
								is_empty := false
							else
								has_elem := true;
							n := n.NextSibling;
						end;

					if is_empty and not has_elem then
						WriteTagClose(xttStartingAndEnding)
					else
						begin
							WriteTagClose(xttStarting);
							if not is_empty then
								Write(Node.NodeValue);
						end;
				end;

			n := Node.FirstChild;
			while n <> nil do
				begin
					if n.NodeType <> xntAttribute then
						begin
							if n.NodeType = xntText then
								begin
									Write(n.NodeValue);
									Inc(FCount);
								end
							else if n.NodeType = xntCData then
								begin
									WriteCDATA(n.NodeValue);
									Inc(FCount);
								end
							else
								WriteNode(n);
						end;
					n := n.NextSibling;
				end;

			if (Node.NodeType = xntElement) and (not is_empty or has_elem) then
				begin
					if has_elem then
						WriteTag(Node.NodeName, xttEnding)
					else
						WriteTag(Node.NodeName, xttEndingSameLine);
				end;

			if Node.NodeType = xntElement then
				DoElementEnd(Node);
		end;

	Inc(FCount);
end;

{ TXMLStringWriter }

constructor TXMLStringWriter.Create(var AString: string);
begin
	inherited Create;
	FString := @AString;
end;

procedure TXMLStringWriter.WriteBuffer(Buffer: PChar; Count: integer);
var
	Size: integer;
begin
	inherited;
	Size := Length(FString^);
	SetLength(FString^, Size+Count);
	Move(Buffer[0], FString^[Size+1], Count);
end;

{ TXMLStreamWriter }

constructor TXMLStreamWriter.Create(AStream: TStream);
begin
	inherited Create;
	FStream := AStream;
end;

procedure TXMLStreamWriter.WriteBuffer(Buffer: PChar; Count: integer);
begin
	inherited;
	FStream.Write(Buffer[0], Count);
end;

{ TXMLFileWriter }

constructor TXMLFileWriter.Create(AFileName: string);
begin
	inherited Create(TFileStream.Create(AFileName, fmCreate or fmShareDenyWrite));
	FFileName := AFileName;
end;

destructor TXMLFileWriter.Destroy;
begin
	if Stream <> nil then
		Stream.Free;
	inherited;
end;

{ TXMLDoc }

constructor TXMLDoc.Create(AOwner: TComponent);
begin
	inherited Create(AOwner);
	FDocument := TXMLNode.Create(xntDocument);
	FIsEmpty := true;
	FFormatOutput := true;
	FProgressStep := 5;
end;

destructor TXMLDoc.Destroy;
begin
	if FDocument <> nil then
		FDocument.Free;
	inherited;
end;

procedure TXMLDoc.Clear(CreateXmlDecl: boolean);
var
  Node: TXMLNode;
begin
	CheckBusy;
	if FDocument <> nil then
		FDocument.Free;
	FDocument := TXMLNode.Create(xntDocument);
	FIsEmpty := true;
  if CreateXmlDecl then
    begin
      // Ubacujemo fiksno zaglavlje:
      // <?xml version='1.0' encoding='windows-1250'?>
      Node := FDocument.CreateChild(xntXMLDeclaration);
      Node.Attributes['version'].NodeValue := '1.0';
      Node.Attributes['encoding'].NodeValue := 'windows-1250';
    end;
end;

(*
function TXMLDoc.createNode(NodeType: TXMLNodeType; const Name: String): TXMLNode;
begin
	result := TXMLNode.Create(NodeType);
	result.NodeName := Name;
end;

function TXMLDoc.createElement(const Name: String): TXMLNode;
begin
	result := createNode(xntElement, Name);
end;

function TXMLDoc.createAttribute(const Name: String): TXMLNode;
begin
	result := createNode(xntAttribute, Name);
end;
*)

procedure	TXMLDoc.DoReadDocumentStart(Reader: TXMLReader);
begin
	if Assigned(FOnReadDocumentStart) then
		FOnReadDocumentStart(Self, Reader);
end;

procedure	TXMLDoc.DoReadDocumentEnd(Reader: TXMLReader);
begin
	if Assigned(FOnReadDocumentEnd) then
		FOnReadDocumentEnd(Self, Reader);
end;

procedure	TXMLDoc.DoReadNode(Sender: TObject);
begin
	FOnReadNode(Self, TXMLReader(Sender));
end;

procedure	TXMLDoc.DoReadElementStart(Sender: TObject);
begin
	FOnReadElementStart(Self, TXMLReader(Sender));
end;

procedure	TXMLDoc.DoReadElementEnd(Sender: TObject);
begin
	FOnReadElementEnd(Self, TXMLReader(Sender));
end;

procedure	TXMLDoc.DoReadProgress(Sender: TObject);
begin
	FOnReadProgress(Self, TXMLReader(Sender));
end;

procedure	TXMLDoc.DoWriteDocumentStart(Writer: TXMLWriter);
begin
	if Assigned(FOnWriteDocumentStart) then
		FOnWriteDocumentStart(Self, Writer);
end;

procedure	TXMLDoc.DoWriteDocumentEnd(Writer: TXMLWriter);
begin
	if Assigned(FOnWriteDocumentEnd) then
		FOnWriteDocumentEnd(Self, Writer);
end;

procedure	TXMLDoc.DoWriteElementStart(Sender: TObject);
begin
	FOnWriteElementStart(Self, TXMLWriter(Sender));
end;

procedure	TXMLDoc.DoWriteElementEnd(Sender: TObject);
begin
	FOnWriteElementEnd(Self, TXMLWriter(Sender));
end;

procedure	TXMLDoc.DoWriteProgress(Sender: TObject);
begin
	FOnWriteProgress(Self, TXMLWriter(Sender));
end;

procedure TXMLDoc.CheckBusy;
begin
	if Loading then
		raise Exception.Create(Name + ':TXMLDoc is still busy loading document');
	if Saving then
		raise Exception.Create(Name + ':TXMLDoc is still busy saving document');
end;

procedure TXMLDoc.LoadFromReader(r: TXMLReader);
begin
	// Clear will call CheckBusy...
	Clear(false);
	FLoading := true;
	try
		r.PreserveWhiteSpace := FPreserveWhiteSpace;
		r.HTMLReader := FHTMLReader;
		r.MixedContent := FMixedContent;
		r.ProgressStep := FProgressStep;

		if Assigned(FOnReadNode) then r.OnNode := DoReadNode;
		if Assigned(FOnReadElementStart) then r.OnElementStart := DoReadElementStart;
		if Assigned(FOnReadElementEnd) then r.OnElementEnd := DoReadElementEnd;
		if Assigned(FOnReadProgress) then r.OnProgress := DoReadProgress;

		DoReadDocumentStart(r);
		r.ParseDocument(FDocument);
		DoReadDocumentEnd(r);

		FIsEmpty := false;
	finally
		FLoading := false;
		r.Free;
	end;
end;

procedure TXMLDoc.LoadFromString(xmlstring: string);
begin
	LoadFromReader(TXMLStringReader.Create(xmlstring));
end;

procedure TXMLDoc.LoadFromStream(xmlstream: TStream);
begin
	LoadFromReader(TXMLStreamReader.Create(xmlstream));
end;

procedure TXMLDoc.LoadFromFile(xmlfilename: string);
begin
	LoadFromReader(TXMLFileReader.Create(xmlfilename));
end;

procedure TXMLDoc.SaveToWriter(w: TXMLWriter);
begin
	CheckBusy;
	FSaving := true;
	try
		w.FormatOutput := FFormatOutput;
		//w.ProgressStep := FProgressStep;

		if Assigned(FOnWriteElementStart) then w.OnElementStart := DoWriteElementStart;
		if Assigned(FOnWriteElementEnd) then w.OnElementEnd := DoWriteElementEnd;
		if Assigned(FOnWriteProgress) then w.OnProgress := DoWriteProgress;

		w.StartDoc;
		DoWriteDocumentStart(w);

		w.WriteNode(FDocument);
		w.EndDoc;
		DoWriteDocumentEnd(w);
	finally
		FSaving := false;
		w.Free;
	end;
end;

function TXMLDoc.SaveToString: string;
begin
	result := '';
	SaveToWriter(TXMLStringWriter.Create(result));
end;

procedure TXMLDoc.SaveToStream(xmlstream: TStream);
begin
	SaveToWriter(TXMLStreamWriter.Create(xmlstream));
end;

procedure TXMLDoc.SaveToFile(xmlfilename: string);
begin
	SaveToWriter(TXMLFileWriter.Create(xmlfilename));
end;

initialization
	NameCharSet := LetterSet + CombiningCharSet + DigitSet + ExtenderSet
			+ ['.', '-', '_', ':'];
	NonQuotedAttrValueEndCharSet := WhiteSpaceSet + ['>'];
end.

